from __future__ import annotations
import typing as T
import json
from pydantic import BaseModel, Field
from fastapi.encoders import jsonable_encoder
from dgraph_orm import GQLException
from .node import Node, Params


NodeType = T.TypeVar("NodeType", bound=Node)

GetParamsType = T.TypeVar("GetParamsType", bound=Params)
QueryParamsType = T.TypeVar("QueryParamsType", bound=Params)
EdgesType = T.TypeVar("EdgesType", bound=BaseModel)

ResolverType = T.TypeVar("ResolverType", bound="Resolver")

BaseModelType = T.TypeVar("BaseModelType", bound=BaseModel)


def get_fields_and_nested_fields(
    cls: T.Type[BaseModelType], include_typename: bool = True
) -> T.List[str]:
    field_names: T.List[str] = []
    for field_name, model_field in cls.__fields__.items():
        t = model_field.type_
        if issubclass(t, Node):
            field_names.append(
                f"{field_name}{{{','.join(get_fields_and_nested_fields(t))}}}"
            )
        else:
            field_names.append(field_name)
    if include_typename:
        return [*field_names, "__typename"]
    return field_names


class Resolver(BaseModel, T.Generic[NodeType]):
    node: T.ClassVar[T.Type[NodeType]]

    query_params: QueryParamsType
    edges: EdgesType

    def gql_fields_str(self) -> str:
        """This does not include the top level..."""
        # TODO this currently just gets all nested fields -> change to make it get edges intelligently
        """DB node is for when you wrap a resolver but need the db node's fields"""
        n = getattr(self, "db_node", getattr(self, "node"))
        fields = get_fields_and_nested_fields(n)
        for resolver_name in self.edges.__fields__.keys():
            resolver: T.Optional[Resolver] = getattr(self.edges, resolver_name, None)
            if resolver:
                child_gql_str = resolver.params_and_fields()
                fields.append(f"{resolver_name} {child_gql_str}")
        return f'{{ {",".join(fields)} }}'

    def params_and_fields(self) -> str:
        return f"{self.query_params.to_str()}{self.gql_fields_str()}"

    def make_get_query_str(self, kwargs_d: dict) -> str:
        kwargs = {k: v for k, v in kwargs_d.items() if v is not None}
        # this is not necessarily true
        if not kwargs:
            pass  # for agg queries (for one)
            # raise GQLException(
            #     f".get requires one field to be given of {list(kwargs_d.keys())}"
            # )
        inner_params = ",".join(
            [
                f"{field_name}: {json.dumps(jsonable_encoder(val))}"
                for field_name, val in kwargs.items()
            ]
        )
        s = f"{{ {self.get_query_name()}({inner_params}){self.gql_fields_str()} }}"
        print(s)
        return s

    def make_get_query_str_via_query(self, kwargs_d: dict) -> str:
        kwargs = {k: v for k, v in kwargs_d.items() if v is not None}
        inner_params = ",".join(
            [
                f"{field_name}: {{eq: {json.dumps(jsonable_encoder(val))} }}"
                for field_name, val in kwargs.items()
            ]
        )
        s = f"{{ {self.query_query_name()}(filter: {{{inner_params}}}){self.gql_fields_str()} }}"
        print(s)
        return s

    @classmethod
    def get_query_name(cls) -> str:
        function_name = cls.node.GQL.get_function_name
        if not function_name:
            raise GQLException(f"No get function found for {cls.node.GQL.typename}!")
        return function_name

    @classmethod
    def query_query_name(cls) -> str:
        function_name = cls.node.GQL.query_function_name
        if not function_name:
            raise GQLException(f"No query function found for {cls.node.GQL.typename}!")
        return function_name

    def make_query_query_str(self) -> str:
        s = f"{{ {self.query_query_name()}{self.params_and_fields()} }}"
        print(s)
        return s

    def make_add_query_str(self) -> str:
        params_and_fields = self.params_and_fields()
        if self.node.GQL.payload_node_name:
            s = f"{{ {self.node.GQL.payload_node_name}{params_and_fields} }}"
        else:
            s = f"{params_and_fields}"
        return s

    def make_delete_query_str(self) -> str:
        fields = self.gql_fields_str()
        if self.node.GQL.payload_node_name:
            s = f"{{ {self.node.GQL.payload_node_name}{fields} }}"
        else:
            s = f"{fields}"
        return s

    def make_add_mutation_str(self, include_upsert: bool) -> str:
        typename = self.node.GQL.typename
        add_model_name = self.node.GQL.add_model.__name__
        add_function_name = self.node.GQL.add_function_name
        if not (self.node.GQL.add_model and add_function_name):
            raise GQLException("Add model or add function is missing")
        upsert_str = f", $upsert: Boolean"
        inner_upsert_str = f", upsert: $upsert"
        if include_upsert is False:
            upsert_str = ""
            inner_upsert_str = ""
        add_param_name = self.node.GQL.add_parameter_name
        s = f"""
            mutation Add{typename}(${add_param_name}: [{add_model_name}!]!{upsert_str}) {{
                {add_function_name}({add_param_name}: ${add_param_name}{inner_upsert_str}) {self.make_add_query_str()}
            }}
        """
        return s

    def make_update_mutation_str(self) -> str:
        typename = self.node.GQL.typename
        patch_model_name = self.node.GQL.patch_model.__name__
        update_function_name = self.node.GQL.update_function_name
        if not (self.node.GQL.patch_model and update_function_name):
            raise GQLException("Update model or update function is missing")
        update_param_name = self.node.GQL.update_parameter_name
        s = f"""
            mutation Update{typename}($set: {patch_model_name}, $remove: {patch_model_name}, $filter: {typename}Filter!) {{
                {update_function_name}({update_param_name}: {{filter: $filter, set: $set, remove: $remove}}) {self.make_add_query_str()}
            }}
        """
        return s

    def make_update_mutation_str_edgedb(self) -> str:
        typename = self.node.GQL.typename
        update_function_name = self.node.GQL.update_function_name
        if not (self.node.GQL.patch_model and update_function_name):
            raise GQLException("Update model or update function is missing")
        s = f"""
            mutation Update{typename} {{
                {update_function_name}(filter: $filter, data: $data) {self.make_delete_query_str()}
            }}
        """
        return s

    def make_delete_mutation_str(self) -> str:
        typename = self.node.GQL.typename

        delete_function_name = self.node.GQL.delete_function_name
        if not delete_function_name:
            raise GQLException("Delete function is missing")

        s = f"""
            mutation Delete{typename}($filter: {typename}Filter!) {{
                {delete_function_name}(filter: $filter) {self.make_delete_query_str()}
            }}
        """
        return s

    async def query(self) -> T.List[NodeType]:
        s = self.make_query_query_str()
        res = await self.node.GQL.dgraph_client.execute(query_str=s)
        lst: T.List[dict] = res["data"][self.query_query_name()]
        return [self.parse_obj_nested(d) for d in lst]

    def parse_obj_nested(self, gql_d: dict) -> NodeType:
        node: NodeType = self.node.parse_obj(gql_d)
        other_fields = set(gql_d.keys()) - set(node.__fields__.keys()) - {"__typename"}
        for field in other_fields:
            resolver = getattr(self.edges, field, None)
            if not resolver:
                raise GQLException(f"No resolver {field} found!")
            nested_d = gql_d[field]
            value_to_save = nested_d
            if nested_d:
                if (list_or_set := type(nested_d)) in {list, set}:
                    val = list_or_set(resolver.parse_obj_nested(d) for d in nested_d)
                else:
                    val = resolver.parse_obj_nested(nested_d)
                value_to_save = val
            node.cache.add(key=field, resolver=resolver, val=value_to_save, gql_d=gql_d)
        node._used_resolver = self
        node._original_dict = node.dict()
        return node

    async def _get(self, kwargs_d: dict) -> T.Optional[NodeType]:
        if not self.node.GQL.get_function_name:
            return await self._get_via_query(kwargs_d)
        s = self.make_get_query_str(kwargs_d=kwargs_d)
        res = await self.node.GQL.dgraph_client.execute(query_str=s)
        obj = res["data"][self.get_query_name()]
        if obj:
            return self.parse_obj_nested(obj)
        return None

    async def _get_via_query(self, kwargs_d: dict) -> T.Optional[NodeType]:
        s = self.make_get_query_str_via_query(kwargs_d=kwargs_d)  # TODO change
        res = await self.node.GQL.dgraph_client.execute(query_str=s)

        obj = res["data"][self.query_query_name()]
        if not obj:
            return None
        return self.parse_obj_nested(obj[0])

    @staticmethod
    def resolvers_by_typename() -> T.Dict[str, T.Type[Resolver]]:
        d = {}
        subs = Resolver.__subclasses__()
        for sub in subs:
            typename = sub.node.GQL.typename
            if typename in d:
                raise GQLException(
                    f"Two Resolvers share the typename {typename}: ({sub.__name__}, {d[typename].__name__})"
                )
            d[typename] = sub
        return d



from __future__ import annotations
import typing as T
import json
import time
from pydantic import BaseModel, PrivateAttr
from fastapi.encoders import jsonable_encoder
from dgraph_orm import GQLException
from devtools import debug
from dgraph_orm.ACL.client import DgraphClient

from dgraph_orm.utils.gql import (
    replace_query_variables_in_query_str,
    turn_set_remove_filter_ds_to_edgedb_ds,
)


def d_rez(refresh: bool = False, use_stale: bool = False, returns: T.Type = None):
    """The problem w this is it does not do types"""

    def outer(f: T.Callable):
        def inner(
            *args, refresh: bool = refresh, use_stale: bool = use_stale, **kwargs
        ) -> returns:
            resolver = f(*args, **kwargs)
            return args[0].resolve(
                name=f.__name__, resolver=resolver, refresh=refresh, use_stale=use_stale
            )

        return inner

    return outer


class Cache(BaseModel):
    # val: T.Union[Node, T.List[Node], None]
    val: T.Any
    resolver: Resolver
    timestamp: float
    raw_gql: str


class CacheManager(BaseModel):
    cache: T.Dict[str, Cache] = {}

    def remove(self, key: str) -> None:
        if key in self.cache:
            del self.cache[key]

    def add(self, *, key: str, resolver: Resolver, val: T.Any, gql_d: dict) -> None:
        self.cache[key] = Cache(
            val=val,
            resolver=resolver,
            timestamp=time.time(),
            raw_gql=json.dumps(jsonable_encoder(gql_d)),
        )

    def replace(self, key: str, cache: Cache) -> None:
        self.cache[key] = cache

    def get(self, key: str) -> T.Optional[Cache]:
        if key not in self.cache:
            return None
        return self.cache[key]

    def exists(self, key: str) -> bool:
        return key in self.cache

    def get_val(self, key: str) -> T.Optional[T.Union[Node, T.List[Node]]]:
        if c := self.cache[key]:
            return c.val

    def clear(self) -> None:
        self.cache = {}

    def is_empty(self) -> bool:
        return len(self.cache) == 0


from .gql_input import GQLInput


def parse_filter(filter: GQLInput) -> str:
    # print(f"{filter=}")
    return filter.to_gql_str()


def parse_nested_q(field_name: str, nested_q: BaseModel):
    if isinstance(nested_q, GQLInput):
        filter_s = parse_filter(nested_q)
        return f"{field_name}: {{ {filter_s} }}"
    outer_lst: T.List[str] = []
    for key, val in nested_q:
        if val is None:
            continue
        # for order, not filter
        if not isinstance(val, BaseModel):
            outer_lst.append(f"{key}: {val}")
            continue
        val: BaseModel
        inner_lst: T.List[str] = []
        for inner_key, inner_val in val.dict(exclude_none=True).items():
            inner_str = f"{inner_key}: {json.dumps(jsonable_encoder(inner_val))}"
            inner_lst.append(inner_str)
        outer_lst.append(f'{key}: {{ {",".join(inner_lst)} }}')
    return f'{field_name}: {{ {",".join(outer_lst)} }}'


class Params(BaseModel):
    def to_str(self) -> str:
        field_names = self.dict(exclude_none=True).keys()
        inner_params: T.List[str] = []
        for field_name in field_names:
            val = getattr(self, field_name)
            if isinstance(val, BaseModel):
                inner_params.append(parse_nested_q(field_name=field_name, nested_q=val))
            else:
                inner_params.append(
                    f"{field_name}: {json.dumps(jsonable_encoder(val))}"
                )
        if inner_params:
            return f'({",".join(inner_params)})'
        return ""

    class Config:
        validate_assignment = True


NodeModel = T.TypeVar("NodeModel", bound="Node")
ResolverType = T.TypeVar("ResolverType", bound="Resolver")
ResolverTempType = T.TypeVar("ResolverTempType", bound="Resolver")

OtherNodeType = T.TypeVar("OtherNodeType", bound="Node")

GQL_D = T.Dict[str, T.Any]
REL_D_INPUT = T.Dict[str, T.Optional[T.Union[OtherNodeType, T.List[OtherNodeType]]]]


class Node(BaseModel):
    _cache: CacheManager = PrivateAttr(default_factory=CacheManager)
    _used_resolver: ResolverType = PrivateAttr(None)
    _original_dict: dict = PrivateAttr(None)
    _deleted: bool = PrivateAttr(None)

    # id: str

    class Config:
        validate_assignment = True

    class GQL:
        # Rename this GQLConfig or something... just GQL probs
        typename: T.ClassVar[str]
        resolver: T.ClassVar[T.Type[ResolverType]]
        payload_node_name: str

        """ TODO fill the add and patch models in future for updating and adding"""
        add_model: T.Type[BaseModel]
        patch_model: T.Type[BaseModel]
        ref_model: T.Type[BaseModel]

        get_function_name: str = None
        query_function_name: str = None

        add_function_name: str = None
        update_function_name: str = None
        delete_function_name: str = None

        add_parameter_name: str = None
        update_parameter_name: str = None

        url: str
        uid_field_name: str

        dgraph_client: DgraphClient

    @property
    def _uid(self) -> T.Union[str, int]:
        return getattr(self, self.GQL.uid_field_name)

    @property
    def _uid_field_name(self) -> str:
        return self.GQL.uid_field_name

    @property
    def used_resolver(self) -> ResolverType:
        return self._used_resolver

    def __eq__(self, other: Node) -> bool:
        if not isinstance(other, Node):
            return False
        return f"{hash(self)}{self.json()}" == f"{hash(other)}{other.json()}"

    def __hash__(self) -> int:
        return hash(f"{self.__class__.__name__}{self._uid}")

    def to_ref(self, field_names: T.List[str] = None) -> GQL.ref_model:
        if field_names:
            return self.GQL.ref_model.parse_obj(
                {field_name: getattr(self, field_name) for field_name in field_names}
            )
        return self.GQL.ref_model.parse_obj(self.dict())

    @property
    def cache(self) -> CacheManager:
        return self._cache

    @staticmethod
    def nodes_by_typename() -> T.Dict[str, T.Type[Node]]:
        d = {}
        subs = Node.__subclasses__()
        for sub in subs:
            typename = sub.GQL.typename
            if typename in d:
                raise GQLException(
                    f"Two Nodes share the typename {typename}: ({sub.__name__}, {d[typename].__name__})"
                )
            d[typename] = sub
        return d

    def __repr__(self) -> str:
        r = super().__repr__()
        r = f"{r}, cache: {repr(self.cache)}" if not self.cache.is_empty() else r
        return r

    def get_root_resolver(self) -> T.Type[Resolver]:
        return Resolver.resolvers_by_typename()[self.GQL.typename]

    @staticmethod
    def should_use_new_resolver(
        old_r: Resolver, new_r: Resolver, strict: bool = False
    ) -> bool:
        old_r_j = old_r.json()
        new_r_j = new_r.json()
        if old_r_j == new_r_j:
            return False
        if strict:
            return True
        if old_r.json(exclude={"edges"}) != new_r.json(exclude={"edges"}):
            print(
                f'excluding children resolvers here..., {old_r.json(exclude={"edges"})=}, {new_r.json(exclude={"edges"})=}'
            )
            return True
        # now do the same for children
        for child_resolver_name in new_r.edges.__fields__.keys():
            new_child_resolver = getattr(new_r.edges, child_resolver_name)
            if new_child_resolver:
                old_child_resolver = getattr(old_r.edges, child_resolver_name)
                if not old_child_resolver:
                    return True
                if Node.should_use_new_resolver(
                    old_r=old_child_resolver,
                    new_r=new_child_resolver,
                    strict=strict,
                ):
                    return True
        return False

    async def resolve(
        self,
        name: str,
        resolver: T.Optional[ResolverTempType],
        refresh: bool = False,
        strict: bool = False,
        use_stale: bool = False,
    ) -> T.Optional[T.Union[NodeModel, T.Set[NodeModel]]]:
        root_resolver = self.GQL.resolver()
        if not resolver:
            resolver = root_resolver.edges.__fields__[name].type_()
        if refresh:
            self.cache.remove(name)
        # see if the resolvers do not match
        if cache := self.cache.get(name):
            if use_stale:
                return cache.val
            if self.should_use_new_resolver(
                old_r=cache.resolver, new_r=resolver, strict=strict
            ):
                print(
                    f"resolvers are different, removing {name} from cache, old: {cache.resolver=}, new: {resolver=}"
                )
                self.cache.remove(name)
        if not self.cache.exists(name):
            setattr(root_resolver.edges, name, resolver)
            obj = await root_resolver._get(kwargs_d={self._uid_field_name: self._uid})
            self.cache.replace(key=name, cache=obj.cache.get(name))
        return self.cache.get_val(name)

    @classmethod
    @property
    def resolver(cls) -> ResolverType:
        return cls.GQL.resolver()

    def hydrate(self, new_node: NodeModel) -> None:
        """Turns this node into the new node"""
        for field_name in new_node.__fields__.keys():
            new_field = getattr(new_node, field_name)
            old_field = getattr(self, field_name)
            if new_field != old_field:
                setattr(self, field_name, getattr(new_node, field_name))
        for private_attr_name in new_node.__private_attributes__.keys():
            setattr(self, private_attr_name, getattr(new_node, private_attr_name))

    """ADD REFRESH BEFORE ADDING ADD"""

    async def refresh(self: NodeModel) -> None:
        new_node = await self._used_resolver._get({self._uid_field_name: self._uid})
        self.hydrate(new_node=new_node)

    """CRUDS"""

    @staticmethod
    def make_set_and_remove_lists(
        original_list: T.Optional[set], new_list: T.Optional[set]
    ) -> T.Tuple[set, set]:
        # TODO should really ONLY be taking in SETS... fix later?
        # add all values from original list not in new list
        if original_list is None:
            original_list = set()
        if new_list is None:
            new_list = set()
        if type(original_list) is list:
            original_list = set(original_list)
        if type(new_list) is list:
            new_list = set(new_list)
        set_list = new_list - original_list
        remove_list = original_list - new_list
        return set_list, remove_list

    def base_set_remove_d(
        self, to_set: REL_D_INPUT = None, to_remove: REL_D_INPUT = None
    ) -> T.Tuple[GQL_D, GQL_D]:
        set_patch = self.GQL.patch_model()
        remove_patch = self.GQL.patch_model()

        current_d = self.dict()
        original_d = self._original_dict

        unset_key = "_UNSET_12398120"

        for field_name in self.GQL.patch_model.__fields__.keys():
            current_val = current_d.get(field_name, unset_key)
            original_val = original_d.get(field_name, unset_key)
            if current_val != original_val:
                # if list, if values have been removed, add to remove... if values have been added, keep
                if {type(current_val), type(original_val)} & {list, set, tuple}:
                    set_list, remove_list = self.make_set_and_remove_lists(
                        original_list=original_val, new_list=current_val
                    )
                    setattr(set_patch, field_name, set_list)
                    setattr(remove_patch, field_name, remove_list)
                else:
                    if current_val != unset_key:
                        # if current val is not unset -> always set
                        if current_val is not None:
                            setattr(set_patch, field_name, current_val)
                        else:
                            setattr(remove_patch, field_name, current_val)
                    else:
                        # current val is UNSET but original val is not
                        setattr(remove_patch, field_name, original_val)

        set_d: GQL_D = set_patch.dict(exclude_unset=True)
        remove_d: GQL_D = remove_patch.dict(exclude_unset=True)

        # stringify values before you include relationships which should not be stringified -> as long as they are REFS
        # i can see this being a problem if you're adding an embedded obj as a Ref ->
        # will deal w that when it comes later
        set_d = self.objs_to_str(set_d)
        remove_d = self.objs_to_str(remove_d)

        # now add relationship fields
        rel_set_d = self.make_gql_d_from_rel_d_input(rel_d_input=to_set or {})
        rel_remove_d = self.make_gql_d_from_rel_d_input(rel_d_input=to_remove or {})
        set_d.update(rel_set_d)
        remove_d.update(rel_remove_d)

        return set_d, remove_d

    async def update(self, given_resolver: ResolverType = None) -> bool:
        """
        Will usually be overwritten assuming there are relationships
        with this node (fill in to set and to remove with those relationship nodes)
        """
        if not self.GQL.update_function_name:
            raise GQLException(f"No update function given for {self.GQL.typename}")
        return await self._update(
            to_set={}, to_remove={}, given_resolver=given_resolver
        )

    @classmethod
    async def update_many(
        cls: T.Type[NodeModel],
        *,
        nodes: T.List[NodeModel],
        set_d: dict = None,
        remove_d: dict = None,
        given_resolver: ResolverType = None,
    ) -> bool:
        if not set_d and not remove_d:
            return False
        variables = {
            "set": jsonable_encoder(set_d) if set_d else {},
            "remove": jsonable_encoder(remove_d) if remove_d else {},
            "filter": {cls.GQL.uid_field_name: [node._uid for node in nodes]},
        }
        resolver = given_resolver or cls.GQL.resolver()
        query_str = resolver.make_update_mutation_str()
        j = await cls.GQL.dgraph_client.execute(
            query_str=query_str, variables=variables
        )
        node_d_lst = j["data"][f"{cls.GQL.update_function_name}"][
            cls.GQL.payload_node_name
        ]
        if not node_d_lst:
            # raise Exception("No update was registered!!")
            print("NO UPDATE WAS REGISTERED")
            return False

        for og_node, node_d in zip(nodes, node_d_lst):
            new_node = resolver.parse_obj_nested(node_d)
            og_node.hydrate(new_node=new_node)

        return True

    @classmethod
    async def _add(
        cls: T.Type[NodeModel],
        *,
        inputs: T.List[GQL.add_model],
        given_resolver: T.Optional[ResolverType] = None,
        upsert: bool = None,
        use_one_time: bool = False,
        use_query_variables: bool = True,  # TODO default for edgedb
    ) -> T.List[NodeModel]:
        resolver = given_resolver or cls.GQL.resolver()
        query_str = resolver.make_add_mutation_str(include_upsert=upsert is not None)
        exclude_none = False if cls.GQL.payload_node_name else True
        variables_val = jsonable_encoder(
            inputs, exclude_unset=True, exclude_none=exclude_none
        )
        # remove empty lists, like admin_permissions = []
        if not cls.GQL.payload_node_name:
            for vars in variables_val:
                for k, v in list(vars.items()):
                    if v == []:
                        del vars[k]
        # also remove all empty arrays
        variables = {cls.GQL.add_parameter_name: variables_val}
        if not cls.GQL.payload_node_name:
            use_query_variables = False
        if upsert is not None:
            variables["upsert"] = upsert
        # print(f"{variables=}, {query_str=}")
        if not use_query_variables:
            # enums must not have "", so remove them
            from enum import Enum

            for i, input in enumerate(inputs):
                enum_lst = [
                    {k: v} for k, v in input.dict().items() if isinstance(v, Enum)
                ]
                # print(f"{enum_lst=}")
                for e_pair in enum_lst:
                    for k, v in e_pair.items():
                        variables_val[i][k] = "<REP>" + variables_val[i][k] + "<REP>"

            better_query_str = replace_query_variables_in_query_str(
                query_str=query_str, query_variables=variables
            )

            better_query_str = better_query_str.replace('"<REP>', "").replace(
                '<REP>"', ""
            )
            # print(f"{better_query_str=}")
            j = await cls.GQL.dgraph_client.execute(
                query_str=better_query_str, use_one_time=use_one_time
            )
        else:
            j = await cls.GQL.dgraph_client.execute(
                query_str=query_str, variables=variables, use_one_time=use_one_time
            )
        inner_response = j["data"][f"{cls.GQL.add_function_name}"]
        if type(inner_response) is dict and (
            node_d_lst := inner_response.get(cls.GQL.payload_node_name)
        ):
            node_d_lst = node_d_lst
        else:
            node_d_lst = inner_response
        return [resolver.parse_obj_nested(node_d) for node_d in node_d_lst]

    @staticmethod
    def make_gql_d_from_rel_d_input(rel_d_input: REL_D_INPUT) -> GQL_D:
        """
        takes {taught_by: Teacher, is_friends_with: T.List[Student]} ->
        {taught_by: {id: 0x1}, is_friends_with: [{id: 0x2, id: 0x3}] }}
        """
        gql_d_output: GQL_D = {}
        for field_name, node_or_nodes in rel_d_input.items():
            if node_or_nodes is None:
                continue
            if type(node_or_nodes) in [set, list, tuple]:
                val = [{n._uid_field_name: n._uid} for n in node_or_nodes]
            else:
                node_or_nodes: NodeModel
                val = {node_or_nodes._uid_field_name: node_or_nodes._uid}
            gql_d_output[field_name] = val
        return gql_d_output

    @staticmethod
    def objs_to_str(obj_d: T.Dict[str, T.Any]) -> T.Dict[str, T.Any]:
        if not obj_d:
            return obj_d
        cleaned_set_d = {}
        for key, val in obj_d.items():
            if isinstance(val, BaseModel):
                val = val.json()
            if type(val) is dict:
                val = json.dumps(jsonable_encoder(val))
            cleaned_set_d[key] = val
        return cleaned_set_d

    async def _update(
        self,
        *,
        given_resolver: T.Optional[ResolverType] = None,
        to_set: T.Optional[REL_D_INPUT] = None,
        to_remove: T.Optional[REL_D_INPUT] = None,
        print_update_d: bool = False,
    ) -> bool:
        set_d, remove_d = self.base_set_remove_d(to_set=to_set, to_remove=to_remove)
        # turn base models or dicts into strs
        """i moved the functions below to inside the function above. This is because relationships should 
        not be turned to string, but nested values should be"""
        # set_d = self.objs_to_str(set_d)
        # remove_d = self.objs_to_str(remove_d)
        return await self.update_from_set_remove_ds(
            set_d=set_d,
            remove_d=remove_d,
            given_resolver=given_resolver,
            print_update_d=print_update_d,
        )

    async def update_from_set_remove_ds(
        self,
        *,
        given_resolver: T.Optional[ResolverType] = None,
        set_d: GQL_D,
        remove_d: GQL_D,
        print_update_d: bool = False,
        use_query_variables: bool = True,  # TODO default for edgedb
    ) -> bool:
        if not self.GQL.payload_node_name:
            use_query_variables = False
        """In future, use args from actual update func from gql.schema"""
        if not set_d and not remove_d:
            print("NOTHING TO UPDATE!")
            return False

        variables = {
            "set": jsonable_encoder(set_d),
            "remove": jsonable_encoder(remove_d),
            "filter": {self._uid_field_name: self._uid},
        }

        if print_update_d:
            debug(variables)

        resolver = given_resolver or self._used_resolver
        query_str = resolver.make_update_mutation_str()
        # print(f"{query_str=}, {variables=}")
        if not use_query_variables:
            query_str = resolver.make_update_mutation_str_edgedb()
            edgedb_variables = turn_set_remove_filter_ds_to_edgedb_ds(variables)
            # debug(edgedb_variables)

            better_query_str = replace_query_variables_in_query_str(
                query_str=query_str, query_variables=edgedb_variables
            )
            print(f"{better_query_str=}")
            j = await self.GQL.dgraph_client.execute(query_str=better_query_str)
        else:
            j = await self.GQL.dgraph_client.execute(
                query_str=query_str, variables=variables
            )
        inner_response = j["data"][f"{self.GQL.update_function_name}"]
        if type(inner_response) is dict and (
            node_d_lst := inner_response.get(self.GQL.payload_node_name)
        ):
            node_d_lst = node_d_lst
        else:
            node_d_lst = inner_response
        if not node_d_lst:
            # raise Exception("No update was registered!!")
            print("NO UPDATE WAS REGISTERED")
            return False

        node_d = node_d_lst[0]
        new_node = resolver.parse_obj_nested(node_d)
        self.hydrate(new_node=new_node)

        return True

    async def delete(
        self,
        given_resolver: ResolverType = None,
        use_query_variables: bool = True,  # TODO default for edgedb
    ) -> bool:
        if not self.GQL.payload_node_name:
            use_query_variables = False
        """In future, use actual GQL.schema"""
        if not self.GQL.delete_function_name:
            raise GQLException(f"No delete function given for {self.GQL.typename}")
        resolver = given_resolver or self._used_resolver
        query_str = resolver.make_delete_mutation_str()
        variables = {"filter": {self._uid_field_name: self._uid}}
        if not self.GQL.payload_node_name:  # edgedb
            variables = {"filter": {"id": {"eq": self._uid}}}
        if not use_query_variables:
            better_query_str = replace_query_variables_in_query_str(
                query_str=query_str, query_variables=variables
            )
            print(f"{better_query_str=}")
            j = await self.GQL.dgraph_client.execute(query_str=better_query_str)
        else:
            j = await self.GQL.dgraph_client.execute(
                query_str=query_str, variables=variables
            )

        inner_response = j["data"][f"{self.GQL.delete_function_name}"]
        if type(inner_response) is dict and (
            node_d_lst := inner_response.get(self.GQL.payload_node_name)
        ):
            node_d_lst = node_d_lst
        else:
            node_d_lst = inner_response
        if not node_d_lst:
            # raise Exception("No update was registered!!")
            print("NO DELETE WAS REGISTERED")
            return False
        if len(node_d_lst) > 1:
            print("MANY WERE DELETED, NOT JUST ONE!")
        node_d = node_d_lst[0]
        node = resolver.parse_obj_nested(node_d)
        node._deleted = True
        self.hydrate(new_node=node)
        return True

    @classmethod
    async def delete_many(
        cls: T.Type[NodeModel],
        *,
        nodes: T.List[NodeModel],
        given_resolver: ResolverType = None,
    ) -> bool:
        """As of now this will not work with EdgeDB"""
        if not cls.GQL.delete_function_name:
            raise GQLException(f"No delete function given for {cls.GQL.typename}")
        resolver = given_resolver or cls.GQL.resolver()
        query_str = resolver.make_delete_mutation_str()
        variables = {
            "filter": {cls.GQL.uid_field_name: [node._uid for node in nodes]},
        }
        j = await cls.GQL.dgraph_client.execute(
            query_str=query_str, variables=variables
        )
        node_d_lst = j["data"][f"{cls.GQL.delete_function_name}"][
            cls.GQL.payload_node_name
        ]
        if not node_d_lst:
            # raise Exception("No update was registered!!")
            print("NO DELETE WAS REGISTERED")
            return False

        for og_node, node_d in zip(nodes, node_d_lst):
            new_node = resolver.parse_obj_nested(node_d)
            new_node._deleted = True
            og_node.hydrate(new_node=new_node)
        return True


from .resolver import Resolver

Cache.update_forward_refs()



import typing as T
import os
import time
import httpx
from httpx import Response
from . import GQLException


class TokenExpiredException(Exception):
    pass


class ThrottleException(Exception):
    pass


class IOTimeoutException(Exception):
    pass


TIMEOUT = float(os.getenv("GQL_CLIENT_TIMEOUT", "25"))
print(f"{TIMEOUT=}")

client_sync = httpx.Client(timeout=TIMEOUT)
client = httpx.AsyncClient(timeout=TIMEOUT)


def check_for_errors(j: dict) -> None:
    if errors := j.get("errors"):
        print(f"{errors=}")
        if "Token is expired" in str(errors):
            raise TokenExpiredException(errors)
        if "Too Many Requests" in str(errors):
            raise ThrottleException(errors)
        if "i/o timeout" in str(errors):
            raise IOTimeoutException(errors)
        raise GQLException(errors)


def finish(
    *, response: Response, query_str: str, should_print: bool, start_time: float
) -> dict:
    try:
        j = response.json()
    except Exception as e:
        message = f".json() did not work for {response=}"
        print(message, "-", f"{e=}")
        raise GQLException(message)
    try:
        print(f"took: {(time.time() - start_time) * 1000}")
        if j.get("extensions"):
            print(
                f'took internal: {int(j["extensions"]["tracing"]["duration"]) / (10 ** 6)}'
            )
        else:
            if os.getenv("USING_EDGEDB", None):
                print("NO EXT BUT EDGEDB")
            else:
                print("NO EXT and not edgedb")
    except Exception as e:
        print(f"ERRORED with {e=} IN FINISH, {j=}")
    if should_print:
        print(f"{query_str=}, {j=}")
    check_for_errors(j)
    if "data" not in j:
        raise GQLException(f"data not in j!, {j=}, {query_str=}")
    return j


"""
DEP
async def gql(
    url: str,
    query_str: str,
    variables: dict = None,
    headers: dict = None,
    should_print: bool = False,
    use_one_time: bool = False,
) -> dict:
    start = time.time()
    headers = headers or {}
    json = {"query": query_str, "variables": variables or {}}
    if use_one_time:
        async with httpx.AsyncClient(timeout=TIMEOUT) as onetime_client:
            response = await onetime_client.post(url=url, json=json, headers=headers)
    else:
        response = await client.post(url=url, json=json, headers=headers)
    return finish(
        response=response,
        query_str=query_str,
        should_print=should_print,
        start_time=start,
    )
"""



from __future__ import annotations
import keyword
import typing as T
from pathlib import Path
import json
import re
from black import format_str, FileMode
import graphql
from pydantic import BaseModel

NODE_CONFIG_NAME = "GQL"
ORM_MODEL_NAME = "GQLInput"
NODE_MODEL_NAME = "Node"
DEFAULT_INDENT = "    "


def imports() -> str:
    lines = [
        "from __future__ import annotations",
        "from enum import Enum",
        "from datetime import datetime, date",
        "from pydantic import BaseModel, Field",
        f"from dgraph_orm import {ORM_MODEL_NAME}, Node, GQLException",
        f"from dgraph_orm.ACL.client import DgraphClient",
        "from dgraph_orm.resolver import Params, Resolver",
        "from typing import Optional, Set, Type, ClassVar, List, Union",
    ]
    return "\n".join(lines)


scalar_d = {
    "String": "str",
    "Int": "int",
    "Float": "float",
    "Boolean": "bool",
    "ID": "str",
    "Int64": "int",
    "DateTime": "datetime",
    # this is for edgedb
    "Bigint": "int",  # ? should be float maybe
    "Decimal": "float",
    # "JSON": "json",
    "JSON": "list",  # could also be dict, for now do not use this, just use
}


def create_payload_node_name(node_name: str) -> str:
    return node_name[0].lower() + node_name[1:]


class NodeConfig(BaseModel):
    node_name: str

    resolver_name: str = None

    # object names
    add_model: str = None
    patch_model: str = None
    ref_model: str = None

    # queries
    get_function_name: str = None
    query_function_name: str = None
    payload_node_name: str = None

    # mutations
    add_function_name: str = None
    update_function_name: str = None
    delete_function_name: str = None

    url: str
    uid_field_name: str

    # not used in GQL
    node_path: str = None

    # for edgedb until it adds get method
    unique_fields: T.List[str] = None

    # for CRUD parameter names -> in future, read this from the introspection
    add_parameter_name: str = "input"
    update_parameter_name: str = "input"

    @classmethod
    def default_from_node_name(cls, node_name: str, url: str) -> NodeConfig:
        return cls(
            node_name=node_name,
            resolver_name=f"{node_name}Resolver",
            get_function_name=f"get{node_name}",
            query_function_name=f"query{node_name}",
            payload_node_name=f"{create_payload_node_name(node_name)}",
            add_function_name=f"add{node_name}",
            update_function_name=f"update{node_name}",
            delete_function_name=f"delete{node_name}",
            add_model=f"Add{node_name}Input",
            patch_model=f"{node_name}Patch",
            ref_model=f"{node_name}Ref",
            url=url,
            uid_field_name="id",
        )

    def create_aggregate(
        self, postfix: str = None, get_function_name: str = None
    ) -> NodeConfig:
        postfix = "AggregateResult" if postfix is None else postfix
        get_function_name = (
            f"aggregate{self.node_name}"
            if get_function_name is None
            else get_function_name
        )
        node_name = f"{self.node_name}{postfix}"
        return self.__class__(
            node_name=node_name,
            url=self.url,
            get_function_name=get_function_name,
            uid_field_name="count",
        )

    @staticmethod
    def str_or_none(s: T.Optional[str], quotes: bool = False) -> str:
        if s is None:
            return "None"
        if quotes:
            s = f'"{s}"'
        return s

    def to_config_class(self) -> str:
        s = f"""
class {NODE_CONFIG_NAME}:
    typename = "{self.node_name}"
    payload_node_name = {self.str_or_none(self.payload_node_name, quotes=True)}
    resolver: Type[{self.str_or_none(self.resolver_name)}]

    # models
    add_model: Type[{self.str_or_none(self.add_model)}] = {self.str_or_none(self.add_model)}
    patch_model: Type[{self.str_or_none(self.patch_model)}] = {self.str_or_none(self.patch_model)}
    ref_model: Type[{self.str_or_none(self.ref_model)}] = {self.str_or_none(self.ref_model)}

    # functions
    get_function_name: str = {self.str_or_none(self.get_function_name, quotes=True)}
    query_function_name: str = {self.str_or_none(self.query_function_name, quotes=True)}

    add_function_name: str = {self.str_or_none(self.add_function_name, quotes=True)}
    update_function_name: str = {self.str_or_none(self.update_function_name, quotes=True)}
    delete_function_name: str = {self.str_or_none(self.delete_function_name, quotes=True)}

    url: str = "{self.url}"
    uid_field_name: str = "{self.uid_field_name}"
    
    dgraph_client: DgraphClient = client
    
    add_parameter_name: str = "{self.add_parameter_name}"
    update_parameter_name: str = "{self.update_parameter_name}"
        """

        return s


def indent_lines(s: str, indent: str = DEFAULT_INDENT) -> str:
    chunks = s.split("\n")
    return indent + f"\n{indent}".join(chunks)


GQL_Type = T.TypeVar("GQL_Type", bound=graphql.GraphQLType)


class PythonType(BaseModel):
    pure_value: str
    is_pure_value_required: bool
    is_list: bool
    is_list_required: bool


def _parse_to_python(
    gql_type: T.Union[
        graphql.GraphQLList,
        graphql.GraphQLEnumType,
        graphql.GraphQLScalarType,
        graphql.GraphQLNonNull,
        graphql.GraphQLObjectType,
    ],
    first: bool = False,
) -> str:
    if first:
        if not isinstance(gql_type, graphql.GraphQLNonNull):
            return f"Optional[{_parse_to_python(gql_type)}]"
    if isinstance(gql_type, graphql.GraphQLNonNull):
        return _parse_to_python(gql_type.of_type)

    if isinstance(gql_type, graphql.GraphQLList):
        prefix = "List["
        postfix = "]"
        if not isinstance(gql_type.of_type, graphql.GraphQLNonNull):
            prefix += "Optional["
            postfix += "]"
        return f"{prefix}{_parse_to_python(gql_type.of_type)}{postfix}"

    if isinstance(gql_type, graphql.GraphQLScalarType):
        return scalar_d[str(gql_type)]

    s = str(gql_type)

    return s


def parse_to_python(
    gql_type: T.Union[
        graphql.GraphQLList,
        graphql.GraphQLEnumType,
        graphql.GraphQLScalarType,
        graphql.GraphQLNonNull,
        graphql.GraphQLObjectType,
    ],
    remove_optional_list: bool = False,
) -> str:
    s = _parse_to_python(gql_type=gql_type, first=True)
    optional_key = "Optional[List["
    if remove_optional_list and optional_key in s:
        s = s.replace(optional_key, "List[")[:-1]
    return s


def get_pure_type(gql_type: GQL_Type) -> T.Type[GQL_Type]:
    while hasattr(gql_type, "of_type"):
        gql_type = gql_type.of_type
    return gql_type


def python_line_from_gql_field(
    field_name: str,
    gql_field: T.Union[
        graphql.GraphQLField, graphql.GraphQLInputField, graphql.GraphQLArgument
    ],
    is_immutable: bool = False,
) -> str:
    python_type_str = parse_to_python(gql_type=gql_field.type)

    field_value_str = "Field("
    inner_value_type = "..."

    # check if optional
    if python_type_str.startswith("Optional["):
        inner_value_type = "None"

    if field_name in keyword.kwlist:
        inner_value_type += f', alias="{field_name}"'
        field_name = f"{field_name}_"

    if is_immutable:
        inner_value_type += ", allow_mutation=False"

    field_value_str = f"{field_value_str}{inner_value_type})"
    s = f"{field_name}: {python_type_str} = {field_value_str}"

    return s


def params_from_gql_arg(field_name: str, gql_arg: graphql.GraphQLArgument) -> str:
    python_type_str = parse_to_python(gql_type=gql_arg.type)

    default_value = None
    # check if optional
    if python_type_str.startswith("Optional["):
        default_value = "None"

    default_value_str = "" if default_value is None else f" = {default_value}"

    return f"{field_name}: {python_type_str}{default_value_str}"


FilterType = T.TypeVar("FilterType", bound=graphql.GraphQLType)


def filter_schema(
    schema: graphql.GraphQLSchema,
    gql_cls: T.Type[FilterType],
    extra_classes: T.List[T.Type[FilterType]] = None,
) -> T.Dict[str, FilterType]:
    gql_clses = [gql_cls]
    if extra_classes:
        gql_clses = [*gql_clses, *extra_classes]
    filtered_d: T.Dict[str, FilterType] = dict()
    for type_name, type_object in schema.type_map.items():
        for gql_cls in gql_clses:
            if isinstance(type_object, gql_cls):
                filtered_d[type_name] = type_object
    return filtered_d


def _build_enums(enum_types_d: T.Dict[str, graphql.GraphQLEnumType]) -> str:
    enum_str_lst: T.List[str] = []
    for type_name, object_type in enum_types_d.items():
        fields_str = "\n".join(
            [f'{val} = "{val}"' for val in object_type.values.keys()]
        )
        s = f"class {type_name}(str, Enum):\n{indent_lines(fields_str)}"
        enum_str_lst.append(s)
    enums_str = "\n".join(enum_str_lst)
    s = format_str(enums_str, mode=FileMode())
    return s


def build_enums(schema: graphql.GraphQLSchema) -> str:
    return _build_enums(filter_schema(schema, graphql.GraphQLEnumType))


def to_pydantic_line(input_name: str) -> str:
    s = f"""
def to_pydantic(self) -> {input_name}:
    return {input_name}.parse_raw(json.dumps(self, cls=EnhancedJSONEncoder))
    """
    return s


def _build_inputs(
    input_types_d: T.Dict[str, graphql.GraphQLInputObjectType],
    for_strawberry: bool = False,
) -> str:
    type_str_lst: T.List[str] = []
    for type_name, object_type in input_types_d.items():
        field_str_lst: T.List[str] = []
        for field_name, gql_field in object_type.fields.items():
            python_line = python_line_from_gql_field(
                field_name=field_name, gql_field=gql_field
            )
            if for_strawberry:
                pure_type = get_pure_type(gql_type=gql_field.type)
                if pure_type not in scalar_d.values():
                    python_line = python_line.replace(
                        str(pure_type), f"Strawberry.{pure_type}"
                    )
                    python_line = (
                        python_line.replace("Field", "strawberry.field")
                        .replace("alias=", "name=")
                        .replace("None", "default=None")
                        .replace("...", "")
                    )
                # ending = "" if "..." in python_line else " = None"
                # python_line = f'{python_line[:python_line.index(" =")]}{ending}'
            field_str_lst.append(python_line)
        if for_strawberry:
            # sort field_str_lst for straw putting required fields first
            field_str_lst.sort(key=lambda _: _.count("="))
        fields_str = "\n".join(field_str_lst)
        class_str = f"class {type_name}({ORM_MODEL_NAME}):"
        if for_strawberry:
            class_str = f"@strawberry.input\nclass {type_name}:"
        type_str = f"{class_str}\n{indent_lines(fields_str)}"
        if for_strawberry:
            type_str += f"\n{indent_lines(to_pydantic_line(input_name=type_name))}"
        type_str_lst.append(type_str)
    types_str = "\n".join(type_str_lst)
    s = format_str(types_str, mode=FileMode())
    return s


def build_inputs(schema: graphql.GraphQLSchema, for_strawberry: bool = False) -> str:
    return _build_inputs(
        filter_schema(schema, graphql.GraphQLInputObjectType),
        for_strawberry=for_strawberry,
    )


def _build_non_input_non_nodes(
    object_types_d: T.Dict[str, graphql.GraphQLObjectType]
) -> str:
    type_str_lst: T.List[str] = []
    for type_name, object_type in object_types_d.items():
        field_str_lst: T.List[str] = []
        for field_name, gql_field in object_type.fields.items():
            python_line = python_line_from_gql_field(
                field_name=field_name, gql_field=gql_field
            )
            field_str_lst.append(python_line)
        fields_str = "\n".join(field_str_lst)
        type_str = f"class {type_name}({ORM_MODEL_NAME}):\n{indent_lines(fields_str)}"
        type_str_lst.append(type_str)
    types_str = "\n".join(type_str_lst)
    s = format_str(types_str, mode=FileMode())
    return s


def build_non_input_non_nodes(
    schema: graphql.GraphQLSchema, node_names: T.List[str]
) -> str:
    d = {
        key: val
        for key, val in filter_schema(schema, graphql.GraphQLObjectType).items()
        if key not in node_names
    }
    return _build_non_input_non_nodes(d)


def _build_nodes(
    node_types_d: T.Dict[str, graphql.GraphQLObjectType],
    node_config_d: T.Dict[str, NodeConfig],
    schema: graphql.GraphQLSchema,
    hydrated: bool = False,
) -> str:
    schema.type_map = {
        key: value
        for key, value in schema.type_map.items()
        if not re.findall(r"[A-Z]\w+_Type", key)
    }
    type_str_lst: T.List[str] = []
    for type_name, object_type in node_types_d.items():
        node_builder = NodeBuilder(
            node_name=type_name,
            gql_object=object_type,
            node_config=node_config_d[type_name],
            schema=schema,
            hydrated=hydrated and node_config_d[type_name].node_path,
        )
        node_resolver_s = node_builder.build_node_and_resolver()
        type_str_lst.append(node_resolver_s)
    types_str = "\n".join(type_str_lst)
    s = format_str(types_str, mode=FileMode())
    return s


def build_nodes(
    schema: graphql.GraphQLSchema,
    node_configs: T.List[NodeConfig],
    hydrated: bool = False,
) -> str:
    nodes_d = filter_schema(schema, graphql.GraphQLObjectType)
    node_names = [n.node_name for n in node_configs]
    filtered_d = {
        node_name: node_object
        for node_name, node_object in nodes_d.items()
        if node_name in node_names
    }
    return _build_nodes(
        filtered_d,
        {n.node_name: n for n in node_configs},
        schema=schema,
        hydrated=hydrated,
    )


class NodeBuilder:
    def __init__(
        self,
        node_name: str,
        node_config: NodeConfig,
        gql_object: graphql.GraphQLObjectType,
        schema: graphql.GraphQLSchema,
        hydrated: bool = False,
    ):
        self.node_name = node_name
        self.node_config = node_config
        self.gql_object = gql_object

        self.fields: T.Dict[str, graphql.GraphQLField] = dict()
        self.edges: T.Dict[str, graphql.GraphQLField] = dict()

        self.create_edges_and_fields(gql_object=gql_object)

        self.schema = schema
        self.query_type = schema.query_type
        self.immutable_fields = []
        if self.node_config.get_function_name:
            self.immutable_fields = self.query_type.fields[
                self.node_config.get_function_name
            ].args.keys()

        self.mutation_type = schema.mutation_type

        self.hydrated = hydrated

    @property
    def resolver_name(self) -> str:
        return f"{self.node_name}Resolver"

    def create_edges_and_fields(self, gql_object: graphql.GraphQLObjectType) -> None:
        for field_name, gql_field in gql_object.fields.items():
            pure_type = get_pure_type(gql_field.type)
            if isinstance(pure_type, graphql.GraphQLObjectType) or isinstance(
                pure_type, graphql.GraphQLInterfaceType
            ):
                self.edges[field_name] = gql_field
            else:
                self.fields[field_name] = gql_field

    def build_edge_functions(self) -> str:
        edge_str_lst: T.List[str] = []
        for edge_name, gql_field in self.edges.items():
            return_type = parse_to_python(
                gql_type=gql_field.type, remove_optional_list=True
            )
            resolver_name = f"{get_pure_type(gql_field.type)}Resolver"
            s = f"""
async def {edge_name}(self, resolver: {resolver_name} = None, refresh: bool = False, use_stale: bool = False) -> {return_type}:
    return await self.resolve(name="{edge_name}", resolver=resolver, refresh=refresh, use_stale=use_stale)
            """
            edge_str_lst.append(s)
        edges_str = "\n".join(edge_str_lst)
        return edges_str

    def build_add(self) -> str:
        if self.node_config.add_function_name is None:
            return ""
        has_upsert = (
            "upsert"
            in self.mutation_type.fields[self.node_config.add_function_name].args
        )
        upsert_str = ", upsert: bool = False"
        if not has_upsert:
            upsert_str = ""
        add_upsert_str = f", upsert=upsert"
        if not has_upsert:
            add_upsert_str = ""
        return f"""
@classmethod
async def add(cls, *, input: {self.node_config.add_model}, resolver: {self.resolver_name} = None{upsert_str}, use_one_time: bool = False) -> {self.node_name}:
    return (await cls.add_many(inputs=[input], resolver=resolver{add_upsert_str}, use_one_time=use_one_time))[0]

@classmethod
async def add_many(cls, *, inputs: List[{self.node_config.add_model}], resolver: {self.resolver_name} = None{upsert_str}, use_one_time: bool = False) -> List[{self.node_name}]:
    return await cls._add(inputs=inputs, given_resolver=resolver, upsert={"upsert" if has_upsert else "None"}, use_one_time=use_one_time)
        """

    def build_update(self) -> str:
        if self.node_config.update_function_name is None:
            return ""

        # now get the patch
        patch_model: graphql.GraphQLInputObjectType = self.schema.type_map[
            self.node_config.patch_model
        ]

        patchable_fields = patch_model.fields.keys()

        arg_str_lst: T.List[str] = []
        to_set_names: T.List[str] = []
        to_remove_names: T.List[str] = []
        for field_name, gql_field in self.edges.items():
            if field_name not in patchable_fields:
                continue
            main_str = ""
            return_type = parse_to_python(
                gql_type=gql_field.type, remove_optional_list=True
            )
            python_type_and_value = f"{return_type} = None"

            set_str = f"{field_name}: {python_type_and_value}"
            to_set_names.append(field_name)
            main_str += set_str

            if not isinstance(gql_field.type, graphql.GraphQLNonNull):
                # if not func.is_required_for_node():
                remove_str = f"remove_{field_name}: {python_type_and_value}"
                main_str += f", {remove_str}"
                to_remove_names.append(field_name)
            arg_str_lst.append(main_str)
        edges_arg_str = ", ".join(arg_str_lst)
        to_set_d_lst: T.List[str] = [f'"{name}": {name}' for name in to_set_names]
        to_remove_d_lst: T.List[str] = [
            f'"{name}": remove_{name}' for name in to_remove_names
        ]
        to_set_d_str = f'{{{", ".join(to_set_d_lst)}}}'
        to_remove_d_str = f'{{{", ".join(to_remove_d_lst)}}}'
        return f"""
async def update(self, resolver: {self.resolver_name} = None, {edges_arg_str}) -> bool:
    return await self._update(given_resolver=resolver, to_set={to_set_d_str}, to_remove={to_remove_d_str})
        """

    def build_fields(self) -> str:
        field_str_lst: T.List[str] = []
        for field_name, gql_field in self.fields.items():
            python_line = python_line_from_gql_field(
                field_name=field_name,
                gql_field=gql_field,
                is_immutable=field_name in self.immutable_fields,
            )
            # change from List to Set for fields on Node
            python_line = python_line.replace("List[", "Set[")
            field_str_lst.append(python_line)
        fields_str = "\n".join(field_str_lst)
        return fields_str

    def build_node_and_resolver(self) -> str:
        node_str = self.build_node()
        resolver_str = self.build_resolver()
        s = f"{self.node_name}.{NODE_CONFIG_NAME}.resolver = {self.node_name}Resolver"
        final_s = "\n".join([node_str, resolver_str, s])
        return format_str(final_s, mode=FileMode())

    def build_node(self) -> str:
        fields_str = self.build_fields()
        if self.hydrated:
            fields_str = ""
        edges_funcs_str = self.build_edge_functions()
        add_func_str = self.build_add()
        update_func_str = self.build_update()
        config_str = self.node_config.to_config_class()
        class_str = f"class {self.node_name}({NODE_MODEL_NAME}):"
        if self.hydrated:
            class_str = f"class {self.node_name}({self.node_name}Hydrated):"
        lines = [
            class_str,
            indent_lines(fields_str),
            indent_lines(edges_funcs_str),
            indent_lines(add_func_str),
            indent_lines(update_func_str),
            indent_lines(config_str),
        ]
        s = "\n".join(lines)
        return format_str(s, mode=FileMode())

    def build_resolver(self) -> str:
        get_field = self.query_type.fields.get(self.node_config.get_function_name, None)
        query_field = self.query_type.fields.get(
            self.node_config.query_function_name, None
        )

        resolver_builder = ResolverBuilder(
            node_name=self.node_name,
            get_field=get_field,
            query_field=query_field,
            edge_field_d=self.edges,
        )

        resolver_str = resolver_builder.build_resolver()
        return resolver_str


def build_model_fields_from_field(field: graphql.GraphQLField) -> str:
    lines: T.List[str] = []
    for arg_name, arg in field.args.items():
        line = python_line_from_gql_field(field_name=arg_name, gql_field=arg)
        lines.append(line)
    if not lines:
        return "pass"
    return "\n".join(lines)


class ResolverBuilder:
    def __init__(
        self,
        node_name: str,
        get_field: T.Optional[graphql.GraphQLField],
        query_field: T.Optional[graphql.GraphQLField],
        edge_field_d: T.Optional[T.Dict[str, graphql.GraphQLField]],
    ):
        self.node_name = node_name
        self.get_field = get_field
        self.query_field = query_field
        self.edge_field_d = edge_field_d

    def build_get_params(self) -> str:
        if not self.get_field:
            fields_str = "pass"
        else:
            fields_str = build_model_fields_from_field(field=self.get_field)
        return f"class {self.node_name}GetParams(Params):\n{indent_lines(fields_str)}"

    def build_query_params(self) -> str:
        if not self.query_field:
            fields_str = "pass"
        else:
            fields_str = build_model_fields_from_field(field=self.query_field)
        return f"class {self.node_name}QueryParams(Params):\n{indent_lines(fields_str)}"

    def build_edges(self) -> str:
        if not self.edge_field_d:
            fields_str = "pass"
        else:
            lines: T.List[str] = []
            for arg_name, arg in self.edge_field_d.items():
                resolver_name = f"{get_pure_type(arg.type)}Resolver"
                line = f"{arg_name}: Optional[{resolver_name}] = None"
                lines.append(line)
            if not lines:
                return "pass"
            fields_str = "\n".join(lines)
        return f"class {self.node_name}Edges(BaseModel):\n{indent_lines(fields_str)}"

    def build_get_functions(self) -> str:
        if not self.get_field:
            if not self.query_field:
                return ""
        if self.get_field:
            params_str = ", ".join(
                [
                    f"{params_from_gql_arg(arg_name, arg)}"
                    for arg_name, arg in self.get_field.args.items()
                ]
            )
        else:
            params_str = "id: str"
        if self.get_field:
            arg_names = self.get_field.args.keys()
        else:
            arg_names = ["id"]
        kwargs_str = ", ".join(f'"{arg_name}": {arg_name}' for arg_name in arg_names)
        kwargs_eq_str = ", ".join(f"{arg_name}={arg_name}" for arg_name in arg_names)
        kwargs_d_str = f"{{{kwargs_str}}}"
        print_str = " and ".join([f"{{{arg_name}=}}" for arg_name in arg_names])
        if self.get_field:
            get = f"""
    async def get(self, {params_str}) -> Optional[{self.node_name}]:
        return await self._get({kwargs_d_str})
            """
        else:
            get = f"""
    async def get(self, {params_str}) -> Optional[{self.node_name}]:
        self.filter(Filter{self.node_name}(id=FilterID(eq=id)))
        results = await self.query()
        if not results:
            return None
        return results[0]
            """
        gerror = f"""
async def gerror(self, {params_str}) -> {self.node_name}:
    node = await self.get({kwargs_eq_str})
    if not node:
        raise GQLException(f"No {self.node_name} with {print_str}")
    return node
        """
        return f"{get}\n{gerror}"

    @property
    def resolver_name(self) -> str:
        return self.node_name + "Resolver"

    def build_query_functions(self) -> str:
        if not self.query_field:
            return ""
        f_str_lst: T.List[str] = []
        for arg_name, arg in self.query_field.args.items():
            python_line = f"{arg_name}: {parse_to_python(gql_type=arg.type)} = None"
            f_str = f"""
def {arg_name}(self, {python_line}, /) -> {self.resolver_name}:
    if {arg_name} is not None:
        if not issubclass({arg_name}.__class__, BaseModel):
            if hasattr({arg_name}, "to_pydantic"):
                {arg_name} = {arg_name}.to_pydantic()
    self.query_params.{arg_name} = {arg_name}
    return self
            """
            f_str_lst.append(f_str.strip())
        return "\n\n".join(f_str_lst)

    def build_edge_functions(self) -> str:
        if not self.edge_field_d:
            return ""
        f_str_lst: T.List[str] = []
        for edge_name, edge_field in self.edge_field_d.items():
            edge_resolver = f"{get_pure_type(edge_field.type)}Resolver"
            f_str = f"""
def {edge_name}(self, _: Optional[{edge_resolver}] = None, /) -> {self.resolver_name}:
    self.edges.{edge_name} = _ or {edge_resolver}()
    return self
            """
            f_str_lst.append(f_str.strip())
        return "\n\n".join(f_str_lst)

    def build_resolver(self) -> str:
        get_params = self.build_get_params()
        query_params = self.build_query_params()
        edges = self.build_edges()

        classes_str_lst = [get_params, query_params, edges]
        classes_str = "\n\n".join(classes_str_lst)
        classes_str = format_str(classes_str, mode=FileMode())

        get_functions = self.build_get_functions()
        query_functions = self.build_query_functions()
        edge_functions = self.build_edge_functions()

        functions_str_lst = [get_functions, query_functions, edge_functions]
        functions_str = "\n\n".join(functions_str_lst)
        functions_str = format_str(functions_str, mode=FileMode())

        edges_str = f"{self.node_name}Edges"
        query_str = f"{self.node_name}QueryParams"

        resolver_header = f"""
class {self.node_name}Resolver(Resolver[{self.node_name}]):
    node: ClassVar[Type[{self.node_name}]] = {self.node_name}
    edges: {edges_str} = Field(default_factory={edges_str})
    query_params: {query_str} = Field(default_factory={query_str}) 
        """

        classes_with_resolver_header = f"{classes_str}\n{resolver_header}"
        classes_with_resolver_header = format_str(
            classes_with_resolver_header, mode=FileMode()
        )

        full_s = f"{classes_with_resolver_header}\n{indent_lines(functions_str)}"
        full_s = format_str(full_s, mode=FileMode())
        return full_s


def get_class_names_from_str(s: str) -> T.List[str]:
    class_names_pattern = r"class (\w*)\("
    return re.findall(class_names_pattern, s)


def build_strawberry_class(schema: graphql.GraphQLSchema) -> str:
    pre = """
import dataclasses
import json
import strawberry

class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, (datetime, date)):
            return o.isoformat()
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return super().default(o)
    """

    enum_names: T.List[str] = []
    for type_name, gql_type in schema.type_map.items():
        if isinstance(gql_type, graphql.GraphQLEnumType):
            if not type_name.startswith("__"):
                enum_names.append(type_name)
    enums_str = "\n".join([f"{name} = strawberry.enum({name})" for name in enum_names])

    inputs_s = build_inputs(schema=schema, for_strawberry=True)
    lines = [
        pre,
        f"class Strawberry:",
        indent_lines(enums_str),
        indent_lines(inputs_s),
    ]
    s = "\n".join(lines)
    s = format_str(s, mode=FileMode())
    return s


def build_all(
    dgraph_client_str: str,
    edgedb_client_str: str,
    node_path_imports_str,
    schema: graphql.GraphQLSchema,
    node_configs: T.List[NodeConfig],
    include_strawberry: bool = False,
    hydrated: bool = False,
) -> str:
    # order map so that you can track db schema changes
    schema.type_map = {
        k: schema.type_map[k] for k in sorted(list(schema.type_map.keys()))
    }

    type_map = {}
    for key, value in schema.type_map.items():
        if re.findall(r"[A-Z]\w+_Type", key):
            key = key.replace("_Type", "")
            value.name = key
        type_map[key] = value
    schema.type_map = type_map

    enums = build_enums(schema)
    inputs = build_inputs(schema)
    non_input_non_nodes = build_non_input_non_nodes(
        schema, node_names=[n.node_name for n in node_configs]
    )
    nodes = build_nodes(schema, node_configs=node_configs, hydrated=hydrated)

    update_forward_ref_lines = [
        f"{n.node_name}Edges.update_forward_refs()" for n in node_configs
    ]
    update_forward_ref_nodes = [
        f"{n.node_name}.update_forward_refs()" for n in node_configs
    ]

    update_forward_ref_lines_str = "\n".join(update_forward_ref_lines)
    update_forward_ref_nodes_str = "\n".join(update_forward_ref_nodes)

    input_names = get_class_names_from_str(inputs)
    non_input_names = get_class_names_from_str(non_input_non_nodes)

    update_forward_ref_input_str = "\n".join(
        [f"{name}.update_forward_refs()" for name in input_names]
    )
    update_forward_ref_type_str = "\n".join(
        [
            f"{name}.update_forward_refs()"
            for name in non_input_names
            if name not in ["Query", "Mutation"]
        ]
    )

    lines = [
        imports(),
        dgraph_client_str,
        edgedb_client_str,
        node_path_imports_str,
        enums,
        inputs,
        non_input_non_nodes,
        nodes,
        update_forward_ref_input_str,
        update_forward_ref_type_str,
        update_forward_ref_lines_str,
        update_forward_ref_nodes_str,
    ]
    final_s = "\n".join(lines)
    s = format_str(final_s, mode=FileMode())
    if include_strawberry:
        straw_str = build_strawberry_class(schema=schema)
        s += f"\n{straw_str}"
        s = format_str(s, mode=FileMode())
    return s


from gql import gql, Client
from gql.transport.aiohttp import AIOHTTPTransport

from dgraph_orm.ACL.client import DgraphClient


async def schema_from_url(
    url: str,
    api_key: str = None,
    user_id: str = None,
    password: str = None,
    namespace: int = None,
) -> graphql.GraphQLSchema:
    headers = {}
    if api_key:
        client = DgraphClient(client_url=url, api_key=api_key)
        await client.login_into_namespace(
            user_id=user_id, password=password, namespace=namespace
        )
        headers = client.headers
    transport = AIOHTTPTransport(url=url, headers=headers)
    client = Client(transport=transport, fetch_schema_from_transport=True)
    query = gql(
        """
        {
          __schema {
            types {
              name
            }
          }
        }
    """
    )
    await client.execute_async(query)
    return client.schema


def node_configs_from_app_config(app_name: str, app_config: dict) -> T.List[NodeConfig]:
    url = app_config["url"]
    nodes = app_config["nodes"]
    node_configs: T.List[NodeConfig] = []
    for node_name, data in nodes.items():
        if data.get("default", False):
            node_config = NodeConfig.default_from_node_name(node_name, url=url)
        else:
            config_data = data.get("config", {})
            node_config = NodeConfig(node_name=node_name, url=url, **config_data)
        node_config.node_path = data.get("node_path", None)
        node_configs.append(node_config)
        if data.get("aggregate", False):
            node_configs.append(node_config.create_aggregate())
    return node_configs


def filename_safe(s: str) -> str:
    return "".join([c for c in s if c.isalpha() or c.isdigit() or c == " "]).rstrip()


def build_dgraph_client(app_config: dict) -> str:
    client_url, api_key = app_config["url"], app_config.get("api_key")
    first_line = f'client = DgraphClient.from_cloud(client_url="{client_url}", api_key="{api_key}")'
    if "namespace" not in app_config:
        return first_line
    s = f"""
{first_line}
async def login_client() -> None:
    await client.login_into_namespace(
        user_id="{app_config['user_id']}",
        password="{app_config['password']}",
        namespace={app_config['namespace']},
    )
    """
    return format_str(s, mode=FileMode())


def build_edgedb_client(app_config: dict) -> str:
    if "edgedb_host" not in app_config or "edgedb_password" not in app_config:
        return ""
    host, password = app_config.get("edgedb_host"), app_config.get("edgedb_password")
    lines = [
        "import edgedb",
        f'edgedb_client = edgedb.create_async_client(tls_security="insecure", host="{host}", password="{password}")',
    ]
    return "\n".join(lines)


def build_node_path_imports(node_configs: T.List[NodeConfig]) -> str:
    strs: T.List[str] = []
    for node in node_configs:
        if node.node_path:
            strs.append(
                f"from {node.node_path} import {node.node_name} as {node.node_name}Hydrated"
            )
            strs.append(
                f"{node.node_name}Hydrated.GQL.resolver.node = {node.node_name}Hydrated"
            )
    return "\n".join(strs)


async def build_from_app_config(
    app_name: str, app_config: dict, parent_dir: Path, include_strawberry: bool = False
) -> None:
    node_configs = node_configs_from_app_config(
        app_name=app_name, app_config=app_config
    )
    schema = await schema_from_url(
        app_config["url"],
        app_config.get("api_key"),
        app_config.get("user_id"),
        app_config.get("password"),
        app_config.get("namespace"),
    )
    dgraph_client_str = build_dgraph_client(app_config)
    edgedb_client_str = build_edgedb_client(app_config)
    node_path_imports_str = build_node_path_imports(node_configs)
    # build once for models, another for nodes if necessary
    s = build_all(
        dgraph_client_str=dgraph_client_str,
        edgedb_client_str=edgedb_client_str,
        node_path_imports_str="",
        schema=schema,
        node_configs=node_configs,
        include_strawberry=include_strawberry,
    )
    open(parent_dir / f"{filename_safe(app_name)}.py", "w").write(s)
    print("already build all once")
    if node_path_imports_str:
        s = build_all(
            dgraph_client_str=dgraph_client_str,
            edgedb_client_str=edgedb_client_str,
            node_path_imports_str=node_path_imports_str,
            schema=schema,
            node_configs=node_configs,
            include_strawberry=include_strawberry,
            hydrated=True,
        )
        open(parent_dir / f"{filename_safe(app_name)}_hydrated.py", "w").write(s)


async def build_from_config(
    config_path: Path, path_to_write_to: Path, include_strawberry: bool = False
) -> None:
    path_to_write_to.mkdir(parents=True, exist_ok=True)
    config = json.loads(open(config_path, "r").read())
    for app_name, app_config in config.items():
        await build_from_app_config(
            app_name=app_name,
            app_config=app_config,
            parent_dir=path_to_write_to,
            include_strawberry=include_strawberry,
        )


if __name__ == "__main__":
    build_from_config(Path("gql_config.json"), Path("generated"))



# from dgraph_orm.resolver import get_fields_and_nested_fields
from datetime import datetime, timedelta
from zoneinfo import ZoneInfo
from pydantic import BaseModel
import time
from dgraph_orm.vendors.edgedb import Node, Resolver, UpdateOperation, Batch
import uuid
from devtools import debug
import typing as T
import asyncio
from pydantic import EmailStr, conint
from enum import Enum
from dgraph_orm.vendors.edgedb.execute import query as execute_query
from fastapi.encoders import jsonable_encoder
import json
from pathlib import Path

# from dgraph_orm.vendors.edgedb.gen import edgedb_DO as DB

from dgraph_orm.vendors.edgedb.gen import aws as DB


async def test_query():
    # person = Person(name="jon", age=2, id=uuid.uuid4())
    # debug(person)
    user_resolver = DB.UserResolver()
    booking_resolver = DB.BookingResolver().limit(10).order_by("created_at desc")
    debug(user_resolver)
    users = await (
        user_resolver.filter(
            "(.name = <str>$name OR .slug = <str>$slug) AND .created_at < <datetime>$created_at",
            # ".slug = <str>$slug",
            variables={
                "slug": "fernandorojo1",
                "name": "Jeremy Berman",
                "created_at": datetime(
                    2021, 10, 15, 11, 25, tzinfo=ZoneInfo("America/New_York")
                ),
            },
        )
        # .offset(0)
        .order_by(".created_at desc")
        # .limit(1)
        .query()
    )
    debug(users)
    debug(user_resolver.dict())


async def test_get():
    # user = await UserResolver().gerror(slug="jeremyberman")
    user = (
        await DB.PersonResolver()
        # .filter(".name = <str>$name", variables={"name": "huan"})
        # .gerror(id="bb53a967-a161-11ec-9241-f78ab4083468")
        .gerror(slug="jeremy")
    )
    debug(user)

    try:
        user = await DB.UserResolver().gerror(
            not_exclusive="bb53a967-a161-11ec-9241-f78ab4083468"
        )
        debug(user)
    except Exception as e:
        print(e)


async def test_nested():
    booking_resolver = (
        DB.BookingResolver()
        .filter(
            ".created_at > <datetime>$created_at",
            variables={
                "created_at": datetime(2000, 10, 1, tzinfo=ZoneInfo("America/New_York"))
            },
        )
        .limit(10)
        .order_by(".start_time asc")
    )
    # print("DJD", booking_resolver.fields_including_nested_to_return)
    print("DJD", booking_resolver.to_str())

    print("HERE")
    user_resolver = (
        DB.UserResolver()
        .bookings(booking_resolver)
        .filter("exists .bookings and .name > <str>$name", variables={"name": "Juan"})
        # .filter("count(.bookings) > 0")
        .order_by("count(.bookings) desc")
    )
    # debug(user_resolver._query_variables)
    # print("USERRR ", user_resolver.fields_including_nested_to_return)
    print("HI THERE", user_resolver.to_str())
    debug(user_resolver._get_nested_query_variables())
    debug(DB.UserResolver._node)
    users = await user_resolver.limit(2).query()
    debug(users)


async def test_add():
    insert_person = DB.PersonInsert(
        first_name="Jon",
        last_name="Quack",
        tags={"hey", "yaaas"},
        ordered_tags=["coding", "python"],
        slug=random_str(10)
        # age=20,
        # created_at=datetime.now(tz=ZoneInfo("America/New_York")),
        # email=EmailStr("j@gmail.com"),
        # version=21,
    )
    p = await DB.Person.add(insert_person)
    debug(p)


async def raw_add_test():
    q_str = "WITH new_Person := (INSERT Person { slug := <str>$slug, first_name := <str>$first_name, last_name := <str>$last_name, tags := array_unpack(<array<str>>$tags) }) SELECT new_Person {first_name, last_name, tags}"
    variables = {
        "first_name": "J",
        "last_name": "M",
        "tags": list({"code", "python"}),
        "slug": random_str(10),
    }
    res = await execute_query(query_str=q_str, variables=variables, client=DB.client)
    debug(res)


async def test_raw_add_many():
    q_str = "WITH new_Persons := (WITH raw_data := <json>$data, FOR item in json_array_unpack(raw_data) UNION (INSERT Person {slug := <str>item['slug'], first_name := <str>item['first_name'], last_name := <str>item['last_name'], tags := array_unpack(<array<str>>item['tags']), created_at := <datetime>$created_at})) SELECT new_Persons {slug, id, first_name, last_name, tags, created_at}"
    variables = {
        "data": json.dumps(
            jsonable_encoder(
                [
                    {
                        "first_name": "J",
                        "slug": random_str(10),
                        "last_name": "Ber",
                        "tags": {"shee", "coding"},
                        "created_at": datetime.now(tz=ZoneInfo("America/New_York")),
                    },
                    {
                        "first_name": "R",
                        "slug": random_str(10),
                        "last_name": "Boo",
                        "tags": {"bloo"},
                    },
                ]
            )
        ),
        "created_at": datetime.now(tz=ZoneInfo("America/New_York")),
        # "created_at": None,
    }
    res = await execute_query(query_str=q_str, variables=variables, client=DB.client)
    debug(res)


async def test_raw_insert_without_for_loop():
    q_str = """
    WITH
        person_1 := (INSERT Person {first_name := <str>$first_name1, slug := <str>$slug1}),
        person_2 := (INSERT Person {first_name := <str>$first_name2, slug := <str>$slug2}),
        person_3 := (INSERT Person {first_name := <str>$first_name3, slug := <str>$slug3}),
        persons := (person_1 UNION person_2 UNION person_3)
    SELECT persons { id, first_name }
    """
    # q_str = "WITH new_Persons := (UNION ( INSERT Person {first_name := <str>$first_name1}, INSERT Person {first_name := <str>$first_name2} ) ), SELECT new_Persons {id, first_name}"
    variables = {
        "first_name1": "Reg",
        "first_name2": "Joe",
        "first_name3": "Juan",
        "slug1": random_str(10),
        "slug2": random_str(10),
        "slug3": random_str(10),
    }
    res = await execute_query(query_str=q_str, variables=variables, client=DB.client)
    debug(res)


async def test_add_many():
    persons = []
    for i in range(2):
        insert_person = DB.PersonInsert(
            first_name=f"Jon{i}",
            last_name=f"Quack{i}",
            tags={"hey", "yaaas"},
            ordered_tags=["coding", "python"],
            created_at=datetime.now(tz=ZoneInfo("America/New_York")),
            slug=random_str(10),
        )
        persons.append(insert_person)
    people = await DB.Person.add_many(inserts=persons)
    debug(people)


async def test_delete():
    insert_person = DB.PersonInsert(first_name="Juacko", slug=random_str(10))
    person = await DB.Person.add(insert_person)

    db_person = await DB.PersonResolver().gerror(id=person.id)
    await db_person.delete()
    assert await DB.PersonResolver().get(id=db_person.id) is None


async def test_delete_many():
    insert_persons = [
        DB.PersonInsert(first_name="Juacko", slug=random_str(10)),
        DB.PersonInsert(first_name="Juano", slug=random_str(10)),
    ]
    persons = await DB.Person.add_many(insert_persons)
    person_ids = [p.id for p in persons]
    db_persons = (
        await DB.PersonResolver()
        .filter(
            ".id in array_unpack(<array<uuid>>$ids)",
            variables={"ids": person_ids},
        )
        .query()
    )
    debug(db_persons)
    assert len(db_persons) == len(persons)
    await DB.Person.delete_many(db_persons)
    print("JUST DELETED THEM")
    debug(db_persons)
    new_db_persons = (
        await DB.PersonResolver()
        .filter(
            ".id in array_unpack(<array<uuid>>$ids)",
            variables={"ids": person_ids},
        )
        .query()
    )
    debug(new_db_persons)
    assert new_db_persons == []


async def test_update():
    person = (
        await DB.PersonResolver().limit(1).order_by("len(.first_name) asc").query()
    )[0]
    person.first_name = person.first_name + "hi"
    n = person.first_name
    await person.update()
    assert person.first_name == n
    db_person = await DB.PersonResolver().gerror(id=person.id)
    debug(db_person)
    assert db_person.first_name == n


async def test_update_many():
    """Actually just make a BATCH... but this can be later"""
    ...


async def test_link():
    jeremy = (
        await DB.PersonResolver()
        .filter(".first_name = <str>$first_name", {"first_name": "Jeremy"})
        .people_i_follow()
        .query()
    )[0]
    debug(jeremy)
    start = time.time()
    people = await jeremy.people_i_follow()
    took = time.time() - start
    assert took < 0.001
    start = time.time()
    people = await jeremy.people_i_follow(DB.PersonResolver().limit(100))
    took = time.time() - start
    assert took > 0.001
    print("LINK TESTING TOOK", took)
    debug(people)

    people = await jeremy.people_i_follow(DB.PersonResolver().my_followers())
    for p in people:
        s = time.time()
        followers = await p.my_followers()
        took = time.time() - s
        assert took < 0.001
        debug(followers)
        s = time.time()
        followers = await p.my_followers(resolver=DB.PersonResolver().people_i_follow())
        took = time.time() - s
        assert took > 0.001
        for follower in followers:
            ss = time.time()
            pp = await follower.people_i_follow()
            assert time.time() - ss
            debug(pp)


async def test_update_links():
    """will have to use detach -> figure out how to do this smartly"""
    jeremy = (
        await DB.PersonResolver()
        .filter(".first_name = <str>$first_name", {"first_name": "Jeremy"})
        .people_i_follow(DB.PersonResolver().limit(3).order_by(".created_at desc"))
        .my_followers()
        .best_friend()
        .query()
    )[0]
    # debug(jeremy)
    followings = await jeremy.people_i_follow(
        DB.PersonResolver().limit(3).order_by(".created_at desc")
    )
    followers = await jeremy.my_followers()
    debug(followings)
    # debug(followers)
    jeremy.tags.add("winkle")
    pr = (
        DB.PersonResolver()
        .filter(".first_name = <str>$first_name", {"first_name": "Fernando"})
        .limit(1)
        .update_operation(UpdateOperation.REPLACE)
    )
    debug(pr.all_filters_str())
    debug(pr._query_variables)
    jons = await (
        DB.PersonResolver()
        .filter(".first_name = <str>$first_name", {"first_name": "Jon"})
        .limit(5)
        .query()
    )
    await jeremy.update(
        best_friend=pr,
        people_i_follow=DB.Person.to_resolver_many(jons).update_operation(
            UpdateOperation.ADD
        ),
    )
    """
    await jeremy.update(
        set_links_d={
            "best_friend": ":= (SELECT DETACHED Person FILTER .first_name = <str>$first_name LIMIT 1)"
        },
        set_links_variables={"first_name": "Fernando"},
    )
    """
    debug(await jeremy.best_friend())
    debug(await jeremy.people_i_follow())
    # ideal

    """
    await jeremy.update(
        best_friend_replace=".first_name"
        # best friend add
        # best friend remove
    )
    """


async def test_batch():
    batch = Batch(print_status=True, client=DB.client)
    for i in range(100):
        insert_person = DB.PersonInsert(
            first_name=f"JonWHOO{i}",
            last_name=f"QuackERS{i}",
            tags={"hey", "yaaas"},
            slug=random_str(6),
        )
        p = await DB.Person.add(insert_person, batch=batch)
        assert p is None
    first_j = (
        await (
            DB.PersonResolver()
            .limit(1)
            .filter(".first_name = <str>$first_name", {"first_name": "J"})
            .query()
        )
    )[0]
    await first_j.delete(batch=batch)
    j_f_rez = DB.PersonResolver().filter(
        ".first_name in array_unpack(<array<str>>$first_names)",
        variables={"first_names": ["Jeremy", "Fernando"]},
    )

    jeremy_fernando = await j_f_rez.query()
    new_created_ats = []
    for person in jeremy_fernando:
        created_at = datetime.now(tz=ZoneInfo("America/New_York"))
        new_created_ats.append(created_at)
        person.created_at = created_at
        await person.update(batch=batch)
    debug(jeremy_fernando)
    ids = await batch.commit(chunk_size=10)
    debug(ids)
    jeremy_fernando_new = await j_f_rez.query()
    debug(new_created_ats)
    debug([p.created_at] for p in jeremy_fernando_new)
    assert [p.created_at for p in jeremy_fernando_new] == new_created_ats

    assert (await DB.PersonResolver().get(id=first_j.id)) == None


async def test_refresh():
    person = (await DB.PersonResolver().limit(1).query())[0]
    og_first_name = person.first_name
    person.first_name = "Sherbert " + og_first_name
    await person.refresh()
    assert person.first_name == og_first_name


async def test_transaction():
    from dgraph_orm.vendors.edgedb.gen.edgedb_DO import client

    async for tx in client.transaction():
        async with tx:

            first_j = (
                await (
                    DB.PersonResolver()
                    .limit(1)
                    .filter(".first_name = <str>$first_name", {"first_name": "J"})
                    .query(given_client=tx)
                )
            )[0]
            debug(first_j)
            for i in range(2):
                insert_person = DB.PersonInsert(
                    first_name=f"JonWHOO{i}",
                    last_name=f"QuackERS{i}",
                    tags={"hey", "yaaas"},
                    slug=random_str(6),
                )
                p = await DB.Person.add(insert_person, given_client=tx)
                debug(p)


from dgraph_orm.vendors.edgedb.constants import random_str, chunk_list, flatten_list


async def update_all_people():
    """NOT A TEST"""
    people = await DB.PersonResolver().query()
    batch = Batch(print_status=True, client=DB.client)
    for person in people:
        person.slug = f"{person.first_name}-{random_str(5)}"
        await person.update(batch=batch)
    res = await batch.commit(chunk_size=30)
    print(res)


async def test_upsert():
    # jeremy = await PersonResolver().gerror(slug="jeremy")
    # debug(jeremy)
    jeremy = await DB.Person.add(
        insert=DB.PersonInsert(
            first_name="Jeremy",
            slug="jeremy",
            last_name="Berman",
            tags={"Gru"},
            ordered_tags=["Skru"],
        ),
        # return_conflicting_model_on="slug",
        upsert_given_conflict_on="slug",
        # custom_on_conflict_str='UNLESS CONFLICT ON .slug ELSE (UPDATE Person SET {tags += "SPP0"})',
    )
    debug(jeremy)


from dgraph_orm.vendors.edgedb.introspection import (
    introspect_objects,
    introspect_scalars,
)


async def test_introspection():
    object_types = await introspect_objects(client=DB.client)
    debug(object_types)
    scalar_types = await introspect_scalars(client=DB.client)
    debug(scalar_types)


async def test_introspection_functions():
    object_types = await introspect_objects()
    for o in object_types:
        for p in [*o.properties, *o.links]:
            print(p.name, p.type_str)


async def test_generated():
    jeremy = await DB.PersonResolver().people_i_follow().gerror(slug="jeremy")
    debug(jeremy)
    for p in await jeremy.people_i_follow():
        debug(p)


async def test_generated_add():
    jonald_insert = DB.PersonInsert(
        first_name="Jonald",
        slug="jonald",
        tags={"wahoo"},
        best_friend=DB.PersonResolver()
        .filter(".slug = <std::str>$slugJ", {"slugJ": "jeremy"})
        .update_operation(UpdateOperation.REPLACE),
    )
    debug(jonald_insert)
    jonald = await DB.Person.add(
        jonald_insert,
        given_resolver=DB.PersonResolver().best_friend(),
        upsert_given_conflict_on="slug",
    )
    debug(jonald)
    debug(await jonald.best_friend())


async def test_generated_delete():
    jonald = await DB.PersonResolver().gerror(slug="jonald")
    await jonald.delete()


def people_inserts(n: int) -> T.List[DB.PersonInsert]:
    return [
        DB.PersonInsert(
            first_name=f"JonWHOO{i}",
            last_name=f"QuackERS{i}",
            tags={"hey", "yaaas"},
            slug=random_str(6),
            best_friend=DB.PersonResolver().filter(
                ".slug = <str>$slug", {"slug": "jeremy"}
            ),
        )
        for i in range(n)
    ]


async def add_lots_async() -> float:
    proms = [DB.Person.add(p) for p in people_inserts(n=500)]
    start = time.time()
    res = await asyncio.gather(*proms)
    took = (time.time() - start) * 1000
    debug(res)
    print("add lots async took", took, "ms")
    return took


async def add_lots_batch() -> float:
    batch = Batch(client=DB.client)
    [await DB.Person.add(p, batch=batch) for p in people_inserts(n=500)]
    start = time.time()
    await batch.commit(25)
    took = (time.time() - start) * 1000
    print("BATCH took", took, "ms")
    return took


async def test_extra_fields():
    class CalendarBookingType(str, Enum):
        external = "external"
        internal = "internal"

    class CalendarBooking(BaseModel):
        firebase_id: str
        artist_name: str
        # artist_profile_image: T.Optional[Image]
        # artist_cover_image: T.Optional[Image]
        start_time: datetime
        booking_type: CalendarBookingType

        # now for the click in fields
        artist_bio: T.Optional[str] = None
        public_event_description: T.Optional[str] = None
        performance_length_mins: int
        artist_slug: T.Optional[str] = None

    start = datetime.now(tz=ZoneInfo("America/New_York"))
    end = start + timedelta(days=90)
    meta_bookings = (
        await DB.MetaBookingResolver()
        .filter(
            ".is_published = <bool>$is_published AND .venue.slug = <str>$slug AND .start_time >= <datetime>$start AND .start_time <= <datetime>$end",
            {
                "start": start,
                "end": end,
                "slug": "beachcomberresort",
                "is_published": True,
            },
        )
        .order_by(".start_time asc")
        .limit(100)
        .offset(1)
        .extra_fields(
            "is_external := exists [is ExternalBooking].id, "
            "artist_name := [is ExternalBooking].artist_name ?? [is Booking].artist.name, "
            "artist_slug := [is Booking].artist.slug, "
            "cover_photo_external := [is ExternalBooking].cover_image, "
            "cover_image := [is Booking].artist.cover_image, artist_bio := [is Booking].artist.bio"
        )
        .query()
    )
    debug(meta_bookings)
    calendar_bookings: T.List[CalendarBooking] = []
    for meta in meta_bookings:
        calendar_bookings.append(
            CalendarBooking(
                firebase_id=meta.firebase_id,
                artist_name=meta.extra.get("artist_name"),
                start_time=meta.start_time,
                booking_type=CalendarBookingType.external
                if meta.extra.get("is_external")
                else CalendarBookingType.internal,
                artist_bio=meta.extra.get("artist_bio"),
                public_event_description=meta.public_event_description,
                performance_length_mins=meta.performance_length_mins,
                artist_slug=meta.extra.get("artist_slug"),
            )
        )
    debug(calendar_bookings)


async def test_bulk_reads():
    users = await DB.UserResolver().query()
    slugs = [u.slug for u in users]

    def create_proms(given_slugs):
        return [
            DB.UserResolver()
            .artists(DB.ArtistResolver().bookings(DB.BookingResolver().limit(10)))
            .venues(DB.VenueResolver().bookings(DB.BookingResolver().limit(10)))
            .gerror(slug=slug)
            for slug in given_slugs
        ]

    all_users_again = []
    start = time.time()
    for x in range(5):
        users_again = await asyncio.gather(*create_proms(slugs))
        all_users_again.extend(users_again)
    took = time.time() - start
    print(f"bulk reading {took=}")
    print(len(users_again))


async def investigate_bulk_load_bug():
    from dgraph_orm.vendors.edgedb.gen import local as db

    # db = DB

    batch = Batch(client=db.client)
    inserts = [
        db.PersonInsert(first_name="jay", slug=random_str(10)) for _ in range(50)
    ]
    print("starting")
    [await db.Person.add(insert=i, batch=batch) for i in inserts]
    res = await batch.commit()
    debug(res)


async def investigate_bulk_load_custom_code():
    import random
    import string

    def random_str(n: int) -> str:
        return "".join(random.choices(string.ascii_letters + string.digits, k=n))

    insert_strs: T.List[str] = []
    model_names: T.List[str] = []
    n = 50
    for i in range(n):
        model_name = f"model_{i}"
        insert_strs.append(
            f"{model_name}: (insert Person {{ first_name := <str>$first_name_{i}, slug := <str>$slug_{i} }} )"
        )

    union_str = f'models := ({" union ".join(model_names)})'
    insert_str = ",\n".join([*insert_strs, union_str])
    query_str = f"""
with
{insert_str}
select models {{id}}
    """
    first_name_variables = {f"first_name_{i}": random_str(10) for i in range(n)}
    variables = {}


async def using_for():
    n = 10
    query_s = f"""
with
    raw_data := <json>$data,
for item in json_array_unpack(raw_data) union (
    with
        # ordered_tags_var := <array<str>>item['ordered_tags'] if false else <array<str>>$empty_array,
        # ordered_tags := <OPTIONAL array<str>>json_get(item, 'ordered_tags_miss'),
        # ordered_tags := <array<str>>json_get(item, 'blo'),
        # ordered_tags := <array<str>>json_get(item, 'json_miss'),
        first_name_var := <str>item['first_name'] if exists <str>json_get(item, 'first_name') else <str>"notset",
        # last_name := <str>item['last_name'] ?? <str>"jon"
        last_name := <str>json_get(item, 'last_namer'),
        # best_friend_filter := <str>".first_name = <str>"Jeremy""
        # best_friend_filter := filter .slug in array_unpack(<array<str>>json_get(item, 'best_friend_slugs')),
        # filter_name := <str>".slug",
        best_friend := (select detached Person filter .slug = <str>json_get(item, 'best_friend', 'filter') limit <int16>json_get(item, 'best_friend', 'limit'))
    insert Person {{ first_name := first_name_var, slug := <str>item['slug'], last_name := last_name, best_friend := best_friend}} unless conflict on .slug else (update Person set {{ first_name := <str>item['first_name'] }})
) {{id, slug, last_name, ordered_tags, tags, inserted_at, created_at, best_friend: {{first_name}} }}
        """
    data = [
        {
            "first_name": random_str(10),
            "slug": random_str(10),
            # "slug": "EPwaYEelMB",
            # "last_name": "Scooby" if i % 2 == 0 else None,
            "last_name": None,
            "tags": ["hey"],
            # "tags": [],
            # "tags": [],
            "ordered_tags": ["spooky"],
            "best_friend_slugs": ["jeremy"],
            "best_friend": {"slug": "jeremy"},
            # "ordered_tags": None,
            # "ordered_tags": [],
        }
        for i in range(n)
    ]
    # learning -> if supposed to be a set or list or tuple but None, set to the empty versions of those things
    # this works if i can treat None as the same as an empty array or set or tuple
    # OK so NULL === {[], {}, ())... will just have to deal with this
    variables = {
        "data": json.dumps(data),
        # "empty_array": ["hi"],
    }
    res = await execute_query(client=DB.client, query_str=query_s, variables=variables)
    debug(res)


async def test_add_many_new():
    n = 10
    inserts = [DB.PersonInsert(first_name="jay", slug=random_str(10)) for _ in range(n)]
    res = await DB.Person.add_many(inserts=inserts, upsert_given_conflict_on="slug")
    debug(res)


from dgraph_orm.vendors.edgedb import creating_strings


async def creating_strings_testing():
    edge_filter_strs = [
        "best_friend := (select detached Person filter .slug = <str>json_get(item, 'best_friend', 'slug')"
    ]
    insert_str = creating_strings.insert_str_from_cls(
        insert_cls=DB.PersonInsert,
        node_cls=DB.Person,
        edge_filter_strs=edge_filter_strs,
        # upsert_given_conflict_on="slug",
        return_conflicting_model_on="slug",
    )
    # print(insert_str)
    insert = DB.PersonInsert(
        first_name="Juanald",
        slug="juanald",
        last_name=random_str(10),
        best_friend=DB.PersonResolver().filter(
            ".slug = <str>$slug", {"slug": "jeremy"}
        ),
    )
    people = await DB.Person.add_many(
        inserts=[insert],
        # upsert_given_conflict_on="slug",
        return_conflicting_model_on="slug",
    )
    debug(people)
    debug(await people[0].best_friend())


async def testing_with_dog():
    inserts = [
        DB.DogInsert(
            name=f"Spoof{i}wwz",
            # slug=random_str(10),
            slug=random_str(10) if i != 0 else "dog",
            # slug="dog",
        )
        for i in range(100)
    ]
    dogs = await DB.Dog.add_many(
        inserts=inserts,
        # return_conflicting_model_on="slug",
        upsert_given_conflict_on="slug",
    )
    debug(dogs)


async def testing_with_dog_raw():
    working_s = """
with
    raw_data := <json>$data,
for item in <tuple<slug: str, name: str>> json_array_unpack(raw_data) union (
    insert Dog {slug := item.slug, name := item.name}
    unless conflict on .slug else (select Dog)
) { slug, name, id }
    """
    testing_s = """
with
    raw_data := <json>$data,
for item in json_array_unpack(raw_data) union (
    with
        slug := <str>json_get(item, 'slug'),
        name := <str>json_get(item, 'name')
    insert Dog {slug := slug, name := name}
    # unless conflict on .slug else (select Dog)
    unless conflict on .slug else (update Dog set {slug := slug, name := name})
) { slug, name, id }
        """
    dogs = await execute_query(
        client=DB.client,
        query_str=testing_s,
        variables={
            "data": json.dumps(
                [
                    {"slug": "dog", "name": "jonny"},
                    {"slug": "dog00", "name": "jonnyald"},
                ]
            )
        },
    )
    debug(dogs)


async def testing_with_person():
    inserts = people_inserts(10_000)
    # people = await DB.Person.add_many(inserts=inerts)
    proms = []
    for chunk in chunk_list(inserts, chunk_size=10000):
        proms.append(
            DB.Person.add_many(
                chunk,
                edge_filter_strs=[
                    "best_friend := (select detached Person filter .slug = <str>json_get(item, 'best_friend', 'slug'))"
                    # "best_friend := (select detached Person order by .inserted_at asc limit 1)"
                ],
            )
        )
    res = flatten_list(await asyncio.gather(*proms))
    print(len(res))
    # print(len(people))
    if False:
        from dgraph_orm.vendors.edgedb.gen.edgedb_DO import client

        async for tx in client.transaction():
            async with tx:
                for insert in inerts:
                    p = await DB.Person.add(insert, given_client=tx)
                    debug(p)


async def testing_add_many_with_people():
    inserts = people_inserts(100)
    inserts[0].slug = "dog"
    inserts[1].slug = "dog"
    # inserts[2].slug = "dog"
    # inserts[0].slug = "sjYZbHr"
    people = await DB.Person.add_many(
        inserts=inserts,
        # return_conflicting_model_on="slug",
        upsert_given_conflict_on="slug",
    )
    debug(people)


async def main():
    await test_get()
    await test_nested()
    await test_add()
    await raw_add_test()
    await test_raw_add_many()
    await test_raw_insert_without_for_loop()
    await test_delete()
    await test_update()
    await test_link()
    await test_update_links()
    await test_batch()
    await test_refresh()
    await test_transaction()
    await test_upsert()
    #
    # await test_generated_add()
    # await test_generated_delete()

    # await test_extra_fields()
    # await test_bulk_reads()
    # await investigate_bulk_load_bug()
    # await using_for()
    # await test_add_many_new()
    # await creating_strings_testing()
    # await testing_with_dog()
    # await testing_with_dog_raw()
    # await testing_with_person()
    # await testing_add_many_with_people()

    # expensive ones
    # await add_lots_async()
    # await add_lots_batch()

    # buggy ones
    # await test_add_many()
    # await test_delete_many()

    # introspection ones
    # await test_introspection()
    # await test_introspection_functions()


if __name__ == "__main__":
    asyncio.run(main())



import typing as T
from dgraph_orm import Node as DBNode
from pydantic import BaseModel, PrivateAttr

StrawberryType = T.TypeVar("StrawberryType")
NodeType = T.TypeVar("NodeType", bound=DBNode)
PydanticType = T.TypeVar("PydanticType", bound=BaseModel)


class LairException(Exception):
    pass


def pydantic_to_straw(
    gql_cls: T.Type[StrawberryType], pyd_model: PydanticType
) -> StrawberryType:
    if pyd_model is None:
        raise LairException("pyd_model cannot be None")
    # this gets all fields that are not resolvers
    straw_type_defs_d = {
        field.name: field
        for field in gql_cls._type_definition.fields
        if field.base_resolver is None
    }
    fields_to_use = set(straw_type_defs_d.keys()) & set(pyd_model.__fields__.keys())
    d = {}
    for field_name in fields_to_use:
        val = getattr(pyd_model, field_name)
        if isinstance(val, BaseModel):
            straw_type = straw_type_defs_d[field_name].type
            if of_type := getattr(straw_type, "of_type", None):
                straw_type = of_type
            val = pydantic_to_straw(gql_cls=straw_type, pyd_model=val)
        d[field_name] = val
    straw = gql_cls(**d)
    straw._node = pyd_model
    return straw


class Display(T.Generic[NodeType]):
    _node: NodeType = PrivateAttr(None)

    @classmethod
    def from_node(cls: T.Type[StrawberryType], node: NodeType) -> StrawberryType:
        return pydantic_to_straw(gql_cls=cls, pyd_model=node)

    @classmethod
    def from_node_or_none(
        cls: T.Type[StrawberryType], node: T.Optional[NodeType]
    ) -> T.Optional[StrawberryType]:
        if node is None:
            return None
        return cls.from_node(node=node)

    @classmethod
    def from_nodes(
        cls: T.Type[StrawberryType], nodes: T.List[NodeType]
    ) -> T.List[StrawberryType]:
        return [cls.from_node(node) for node in nodes]

    @property
    def node(self) -> NodeType:
        return self._node


DBType = T.TypeVar("DBType", bound=DBNode)
LayerNodeType = T.TypeVar("LayerNodeType", bound=DBNode)

# TODO prob rename this...
class Node(T.Generic[DBType]):
    @classmethod
    def from_db(cls: T.Type[LayerNodeType], model_db: DBType) -> LayerNodeType:
        # TODO make sure this covers everything
        # but must also transfer private fields and cache!
        n = cls(**model_db.dict())
        # for all private fields too
        # ooohhh, this is for the cache!
        for private_field in model_db.__private_attributes__.keys():
            setattr(n, private_field, getattr(model_db, private_field))
        return n

    @classmethod
    def from_db_or_none(
        cls: T.Type[LayerNodeType], model_db: T.Optional[DBType]
    ) -> T.Optional[LayerNodeType]:
        if not model_db:
            return None
        return cls.from_db(model_db)

    @classmethod
    def from_dbs(
        cls: T.Type[LayerNodeType], model_dbs: T.List[DBType]
    ) -> T.List[LayerNodeType]:
        return [cls.from_db(model_db) for model_db in model_dbs]



import typing as T
from enum import Enum
from datetime import datetime
from pydantic import BaseModel
from fastapi.encoders import jsonable_encoder
from dgraph_orm import GQLException

_NULL = "null-12345**792387fsijfoijf"


class GQLInput(BaseModel):
    def to_gql_str(self) -> str:
        str_lst: T.List[str] = []
        for field_name in self.dict(exclude_none=True, by_alias=True):
            val = getattr(self, field_name, None)
            if val is None:
                val = getattr(self, f"{field_name}_")
            if type(val) is set:
                val = list(val)
            if type(val) is list:
                if len(val) == 0:
                    raise GQLException(f"List in filter cannot be empty, {self=}")
                if isinstance(val[0], Enum):
                    v_str = ",".join(f"{v.value}" for v in val)
                    s = f"{field_name}: [{v_str}]"
                    str_lst.append(s)
                elif isinstance(val[0], GQLInput):
                    s = f'{field_name}: {{ {",".join([v.to_gql_str() for v in val])} }}'
                    str_lst.append(s)
                else:
                    s = f"{field_name}: {jsonable_encoder(val)}"
                    str_lst.append(s)
            else:
                if val == _NULL:
                    s = f"{field_name}: null"
                    str_lst.append(s)
                elif isinstance(val, GQLInput):
                    val = val.to_gql_str()
                    s = f"{field_name}: {{ {val} }}"
                    str_lst.append(s)
                elif isinstance(val, Enum):
                    s = f"{field_name}: {val.value}"
                    str_lst.append(s)
                elif type(val) is str:
                    s = f'{field_name}: "{val}"'
                    str_lst.append(s)
                elif type(val) is bool:
                    s = f'{field_name}: {"false" if val is False else "true"}'
                    str_lst.append(s)
                elif type(val) is datetime:
                    s = f'{field_name}: "{val.isoformat()}"'
                    str_lst.append(s)
                else:
                    s = f"{field_name}: {val}"
                    str_lst.append(s)
        final_s = ",".join(str_lst)
        final_s = final_s.replace("'", '"')
        return final_s

    class Config:
        #     use_enum_values = True
        allow_population_by_field_name = True



import json
import re
import typing as T
from enum import Enum
import os
from pathlib import Path
from black import format_str, FileMode
import edgedb
from pydantic import BaseModel
from devtools import debug
from .introspection import (
    introspect_objects,
    introspect_scalars,
    ObjectType,
    ScalarType,
    Link,
    Property,
    Cardinality,
)

PATH_TO_MODULE = "dgraph_orm.vendors.edgedb"
DEFAULT_INDENT = "    "
CONFIG_NAME = "GraphORM"


class GeneratorException(Exception):
    pass


def indent_lines(s: str, indent: str = DEFAULT_INDENT) -> str:
    chunks = s.split("\n")
    return indent + f"\n{indent}".join(chunks)


def imports() -> str:
    lines = [
        "from __future__ import annotations",
        "import typing as T",
        "from enum import Enum",
        "from datetime import datetime, date, timedelta",
        "from uuid import UUID",
        "from decimal import Decimal",
        "from edgedb import RelativeDuration, AsyncIOClient, create_async_client",
        "from pydantic import BaseModel, Field",
        f"from {PATH_TO_MODULE} import Node, Resolver, NodeException, ResolverException, UpdateOperation, Batch",
    ]
    return "\n".join(lines)


async def build_enums(client: edgedb.AsyncIOClient) -> str:
    scalar_types = await introspect_scalars(client)
    enum_strs: T.List[str] = []
    for scalar in scalar_types:
        if not scalar.enum_values:
            continue
        enum_value_strs: T.List[str] = [f'{e} = "{e}"' for e in scalar.enum_values]
        enum_value_str = "\n".join(enum_value_strs)
        s = f"class {scalar.node_name}(str, Enum):\n{indent_lines(enum_value_str)}"
        enum_strs.append(s)
    return "\n".join(enum_strs)


def build_node_link_function_str(link: Link) -> str:
    link_resolver_name = f"{link.target.model_name}Resolver"
    return f"""
async def {link.name}(
    self,
    resolver: {link_resolver_name} = None,
    refresh: bool = False,
    force_use_stale: bool = False,
) -> {link.type_str}:
    return await self.resolve(
        edge_name="{link.name}",
        edge_resolver=resolver or {link_resolver_name}(),
        refresh=refresh,
        force_use_stale=force_use_stale,
    )
    """


def build_resolver_link_function_str(node_resolver_name: str, link: Link) -> str:
    link_resolver_name = f"{link.target.model_name}Resolver"
    return f"""
def {link.name}(self, _: T.Optional[{link_resolver_name}] = None, /) -> {node_resolver_name}:
    if "{link.name}" in self._nested_resolvers:
        raise ResolverException("A resolver for `{link.name}` has already been provided.")
    self._nested_resolvers["{link.name}"] = _ or {link_resolver_name}()
    return self
    """


def build_update_function_str(node_resolver_name: str, links: T.List[Link]) -> str:
    link_strs: T.List[str] = []
    link_names: T.List[str] = []
    for link in links:
        if link.readonly:
            continue
        link_strs.append(
            f"{link.name}: T.Optional[{link.target.model_name}Resolver] = None"
        )
        link_names.append(link.name)
    link_params_str = ", ".join(link_strs)
    return f"""
async def update(
    self,
    given_resolver: {node_resolver_name} = None,
    error_if_no_update: bool = False,
    batch: Batch = None,
    given_client: AsyncIOClient = None,
    {link_params_str}
) -> None:
    set_links_d = {{{", ".join([f'"{link_name}": {link_name}' for link_name in link_names])}}}
    set_links_d = {{key: val for key, val in set_links_d.items() if val is not None}}
    
    return await super().update(
        given_resolver=given_resolver,
        error_if_no_update=error_if_no_update,
        set_links_d=set_links_d,
        batch=batch,
        given_client=given_client
    )
    """


def add_quotes(lst: T.Iterable[str]) -> T.Iterable[str]:
    return [f'"{o}"' for o in lst]


def build_orm_config(
    model_name: str, updatable_fields: T.Set[str], exclusive_fields: T.Set[str]
) -> str:
    return f"""
class {CONFIG_NAME}:
    model_name = "{model_name}"
    client = client
    updatable_fields: T.Set[str] = {{{', '.join(add_quotes(sorted(list(updatable_fields))))}}}
    exclusive_fields: T.Set[str] = {{{', '.join(add_quotes(sorted(list(exclusive_fields))))}}}
    """


def stringify_dict(d: T.Union[T.Dict[str, str], str]) -> str:
    if type(d) is not dict:
        s = f"{d}"
        if type(d) is not bool:
            s = f'"{s}"'
        return s
    inner = [f'"{k}":{stringify_dict(v)}' for k, v in d.items()]
    return f"{{{','.join(inner)}}}"


def edgedb_conversion_type_from_prop(prop: Property) -> str:
    s = prop.target.name
    pattern = r"default::\w+"
    s = re.sub(pattern, "std::str", s)
    return s


def build_node_and_resolver(object_type: ObjectType) -> str:
    # need to sort props and links by required, exclusive, no default, rest
    object_type.properties.sort(
        key=lambda x: f"{x.required}-{x.is_exclusive}-{x.default}",
        reverse=True,
    )
    object_type.links.sort(
        key=lambda x: f"{x.required}-{x.is_exclusive}-{x.default}",
        reverse=True,
    )
    # start with the properties
    node_resolver_name = f"{object_type.node_name}Resolver"
    property_strs: T.List[str] = []
    insert_property_strs: T.List[str] = []
    updatable_fields: T.Set[str] = set()
    exclusive_fields: T.Set[str] = set()

    node_edgedb_conversion_map: T.Dict[str, T.Dict[str, str]] = {}
    insert_edgedb_conversion_map: T.Dict[str, T.Dict[str, str]] = {}

    for prop in object_type.properties:
        conversion_type = edgedb_conversion_type_from_prop(prop)
        node_edgedb_conversion_map[prop.name] = {
            "cast": conversion_type,
            "cardinality": prop.cardinality.value,
            "readonly": prop.readonly,
        }
        if not prop.readonly:
            updatable_fields.add(prop.name)
        if prop.is_exclusive:
            exclusive_fields.add(prop.name)
        default_value_str = "..." if prop.required else "None"
        allow_mutation_str = f"allow_mutation={not prop.readonly}"
        property_strs.append(
            f"{prop.name}: {prop.type_str} = Field({default_value_str}, {allow_mutation_str})"
        )
        # for insert type
        if prop.name != "id":
            insert_edgedb_conversion_map[prop.name] = {
                "cast": conversion_type,
                "cardinality": prop.cardinality.value,
                "readonly": prop.readonly,
            }
            insert_type_str = prop.type_str
            # if required but has default, add optional back
            if prop.required and prop.default:
                insert_type_str = f"T.Optional[{insert_type_str}]"
            default_value_str = (
                " = None" if insert_type_str.startswith("T.Optional[") else ""
            )
            insert_property_strs.append(
                f"{prop.name}: {insert_type_str}{default_value_str}"
            )
    link_function_strs: T.List[str] = []
    resolver_function_strs: T.List[str] = []
    updatable_links: T.Set[str] = set()
    exclusive_links: T.Set[str] = set()
    link_conversion_map: T.Dict[str, T.Dict[str, str]] = {}

    for link in object_type.links:
        if link.name == "__type__":
            continue
        link_conversion_map[link.name] = {
            "cast": link.target.model_name,
            "cardinality": link.cardinality.value,
            "readonly": link.readonly,
            "required": link.required,
        }
        if not link.readonly:
            updatable_links.add(link.name)
        if link.is_exclusive:
            exclusive_links.add(link.name)
        link_function_strs.append(build_node_link_function_str(link))
        resolver_function_strs.append(
            build_resolver_link_function_str(
                node_resolver_name=node_resolver_name, link=link
            )
        )
        # for insert
        insert_resolver_str = f"{link.target.model_name}Resolver"
        if (not link.required) or (link.required and link.default):
            insert_resolver_str = f"T.Optional[{insert_resolver_str}]"
        default_value_str = (
            " = None" if insert_resolver_str.startswith("T.Optional[") else ""
        )
        insert_property_strs.append(
            f"{link.name}: {insert_resolver_str}{default_value_str}"
        )
    orm_config_str = build_orm_config(
        model_name=object_type.node_name,
        updatable_fields={*updatable_fields, *updatable_links},
        exclusive_fields={*exclusive_fields, *exclusive_links},
    )

    insert_model_name = f"{object_type.node_name}Insert"

    # insert type
    insert_inner_str = "\n".join(insert_property_strs)
    insert_conversion_map_str = f"_edgedb_conversion_map: T.ClassVar[T.Dict[str, str]] = {stringify_dict(insert_edgedb_conversion_map)}"
    insert_s = f"class {insert_model_name}(BaseModel):\n{indent_lines(insert_inner_str)}\n\n{indent_lines(insert_conversion_map_str)}"

    # node
    node_properties_str = "\n".join(property_strs)
    node_conversion_map_str = f"_edgedb_conversion_map: T.ClassVar[T.Dict[str, str]] = {stringify_dict(node_edgedb_conversion_map)}"
    insert_link_conversion_map_str = f"_link_conversion_map: T.ClassVar[T.Dict[str, str]] = {stringify_dict(link_conversion_map)}"
    node_link_functions_str = "\n".join(link_function_strs)
    update_function_str = build_update_function_str(
        node_resolver_name=node_resolver_name, links=object_type.links
    )
    node_inner_strs = [
        node_properties_str,
        "\n",
        node_conversion_map_str,
        insert_link_conversion_map_str,
        node_link_functions_str,
        update_function_str,
        orm_config_str,
    ]
    node_inner_str = "\n".join(node_inner_strs)
    node_s = f"class {object_type.node_name}(Node[{insert_model_name}]):\n{indent_lines(node_inner_str)}"

    # resolver
    resolver_properties_str = f"_node = {object_type.node_name}"
    resolver_link_functions_str = "\n".join(resolver_function_strs)
    resolver_inner_strs = [resolver_properties_str, resolver_link_functions_str]
    resolver_inner_str = "\n".join(resolver_inner_strs)
    resolver_s = f"class {node_resolver_name}(Resolver[{object_type.node_name}]):\n{indent_lines(resolver_inner_str)}"

    final_s = (
        f"{object_type.node_name}.{CONFIG_NAME}.resolver_type = {node_resolver_name}"
    )
    return f"{insert_s}\n{node_s}\n{resolver_s}\n{final_s}"


async def build_nodes_and_resolvers(client: edgedb.AsyncIOClient) -> str:
    object_types = await introspect_objects(client)
    node_strs: T.List[str] = []
    for object_type in object_types:
        node_strs.append(build_node_and_resolver(object_type))
    update_forward_refs_str = "\n".join(
        [f"{o.node_name}Insert.update_forward_refs()" for o in object_types]
    )
    nodes_str = "\n".join(node_strs)
    return f"{nodes_str}\n\n{update_forward_refs_str}"


class DBVendor(str, Enum):
    edgedb = "edgedb"


class TLSSecurity(str, Enum):
    insecure = "insecure"


class DBConfig(BaseModel):
    vendor: DBVendor
    host: str
    password: str
    port: int = 5656
    tls_security: TLSSecurity


def build_client(db_config: DBConfig) -> str:
    return f"""
client = create_async_client(
    tls_security="{db_config.tls_security.value}",
    host="{db_config.host}",
    password="{db_config.password}", 
    port={db_config.port}
)
    """


def validate_output_path(path: Path) -> None:
    if not os.path.isdir(path):
        if os.path.isfile(path):
            raise GeneratorException(
                f"output path {path=} must be a directory, not a file."
            )
        if not os.path.exists(path):
            os.makedirs(path)


async def generate(config_path: Path, output_path: Path) -> None:
    validate_output_path(output_path)
    for db_name, db_config_d in json.loads(open(config_path).read()).items():
        db_config = DBConfig.parse_obj(db_config_d)
        client = edgedb.create_async_client(
            tls_security=db_config.tls_security,
            host=db_config.host,
            password=db_config.password,
            port=db_config.port,
        )
        imports_str = imports()
        client_str = build_client(db_config)
        enums_str = await build_enums(client)
        nodes_and_resolvers_str = await build_nodes_and_resolvers(client)

        s = "\n".join([imports_str, client_str, enums_str, nodes_and_resolvers_str])
        s = format_str(s, mode=FileMode())

        open(output_path / f"{db_name}.py", "w").write(s)



import typing as T
import asyncio
from pathlib import Path

from dgraph_orm.vendors.edgedb.generator import generate


async def main():
    await generate(
        config_path=Path("graphorm_config.json"),
        output_path=Path(__file__).parent / "gen",
    )


if __name__ == "__main__":
    asyncio.run(main())



import typing as T
import json
import re
from devtools import debug


def query_variable_to_str(query_variable: T.Union[dict, T.List[dict]]) -> str:
    """
    INPUT(list[dict]|dict): [{"first_name":"Harry", "last_name": "Potter"}, {"first_name":"Ron", "last_name": "Weasley"}]
    OUTPUT (str): [{first_name: "Harry", last_name: "Potter"}, {first_name: "Ron", last_name: "Weasley"}]
    """
    # str everything
    # if root value (does not have a : directly infront of it), keep "". Otherwise, do not. Better yet, if has a ":" infront of it, remove the ""
    # pattern = r'[\w"\s]+:'
    pattern = r'(")([\w]+)(")(:)'
    return re.sub(pattern, r"\2\4", str(json.dumps(query_variable)))


def replace_query_variables_in_query_str(query_str: str, query_variables: dict) -> str:
    pattern = r"(\(\$[\w:\s\[!\]]+\))\s*({)"
    better_query_str = re.sub(pattern, r"\2", query_str)
    for key, val in query_variables.items():
        query_variable = query_variable_to_str(val)
        better_query_str = better_query_str.replace(f"${key}", query_variable)
    # print(f"{better_query_str=}")
    return better_query_str


"""
FROM
query_variables: {"data": [{"first_name":"Harry", "last_name": "Potter"}]}
mutation addPerson($data: [InsertPerson!]!) {
  insert_Person(data: $data) {
    id
  }
}
TO
mutation addPersonNoVar {
  insert_Person(data: [{first_name: "Harry", last_name: "Potter"}]) {
    id
  }
}
"""


def turn_set_remove_filter_ds_to_edgedb_ds(variables: dict) -> dict:
    """
    INPUT
    variables: {
        'set': {
            'title': 'New Titleman',
            'actors': {'id': '7ef21044-9fde-11ec-9407-bfe50cd00551'},
        },
        'remove': {'year': None},
        'filter': {
            'id': '663797f2-9fe0-11ec-99a0-3b9099c2b109',
        },
    }
    OUTPUT
    {
        'data': {
            'title': {'set': 'New Titleman'},
            'year': {'clear': True},
            'actors': {'set': [{'filter': {'id': {'eq': '7ef21044-9fde-11ec-9407-bfe50cd00551'}}}]}
        },
        'filter': {'id': {'eq': '663797f2-9fe0-11ec-99a0-3b9099c2b109'}}
    }
    """

    def turn_id_into_filter_id(input_d: dict) -> dict:
        """
        INPUT
        {
            "title": "New Titleman",
            "actors": {"id": "7ef21044-9fde-11ec-9407-bfe50cd00551"},
        }
        OUTPUT
        {
            "title": "New Titleman",
            "actors": {"filter": {"id": {"eq": "7ef21044-9fde-11ec-9407-bfe50cd00551"}}},
        }
        """
        output_d = {}
        for k, v in input_d.items():
            if type(v) is list:
                if v and type(v[0]) is dict and "id" in v[0]:
                    v = [{"filter": {"id": {"eq": i["id"]}}} for i in v]
            else:
                if type(v) is dict:
                    if m_id := v.get("id"):
                        v = {"filter": {"id": {"eq": m_id}}}
            output_d[k] = v
        return output_d

    new_d = {"data": {}}
    variables["set"] = turn_id_into_filter_id(input_d=variables["set"])
    variables["remove"] = turn_id_into_filter_id(input_d=variables["remove"])
    for key, val in variables["set"].items():
        if key not in new_d["data"]:
            new_d["data"][key] = {}
        existing_d = new_d["data"][key]
        existing_d["set"] = val
    for key, val in variables["remove"].items():
        if key not in new_d["data"]:
            new_d["data"][key] = {}
        existing_d = new_d["data"][key]
        if type(val) == dict and "filter" in val:
            existing_d["remove"] = val
        elif type(val) == list and "filter" in val[0]:
            existing_d["remove"] = val
        else:
            existing_d["clear"] = True
    model_id = variables["filter"]["id"]
    new_d["filter"] = {"id": {"eq": model_id}}
    return new_d


if __name__ == "__main__":
    """
    res = query_variable_to_str(
        query_variable=[
            {"first_name": "Harry", "last_name": "Potter"},
            {"first_name": "Ron", "last_name": "Weasley", "inner": {"hey": "yes"}},
        ]
    )
    print(f"{res=}")
    """
    variables = {
        "set": {
            "title": "New Titleman",
            "actors": [{"id": "7ef21044-9fde-11ec-9407-bfe50cd00551"}, {"id": "123"}],
        },
        "remove": {"year": None, "actors": {"id": "034"}},
        "filter": {
            "id": "663797f2-9fe0-11ec-99a0-3b9099c2b109",
        },
    }
    res = turn_set_remove_filter_ds_to_edgedb_ds(variables)
    debug(res)
